from __future__ import absolute_import, print_function

import datetime
import os
import shutil
import tempfile
import textwrap

import mock

from egginst._compat import assertCountEqual
from egginst.tests.common import mkdtemp
from egginst.vendor.six.moves import unittest

from enstaller.config import Configuration
from enstaller.tests.common import (FAKE_MD5, FAKE_SIZE, PY_VER,
                                    FakeOptions,
                                    create_prefix_with_eggs,
                                    create_repositories,
                                    dummy_installed_package_factory,
                                    dummy_repository_package_factory,
                                    mock_print)

from ..commands import (info_option, install_from_requirements, update_all,
                        whats_new)


class TestInfoStrings(unittest.TestCase):
    def test_info_option(self):
        self.maxDiff = None

        # Given
        mtime = 0.0
        r_output = textwrap.dedent("""\
        Package: enstaller

        Version: 4.6.2-1
            Product: commercial
            Available: True
            Python version: {2}
            Store location: {3}
            Last modified: {4}
            MD5: {0}
            Size: {1}
            Requirements: None
        Version: 4.6.3-1
            Product: commercial
            Available: True
            Python version: {2}
            Store location: {3}
            Last modified: {4}
            MD5: {0}
            Size: {1}
            Requirements: None
        """.format(FAKE_MD5, FAKE_SIZE, PY_VER, "",
                   datetime.datetime.fromtimestamp(mtime)))

        entries = [dummy_repository_package_factory("enstaller", "4.6.2", 1),
                   dummy_repository_package_factory("enstaller", "4.6.3", 1)]

        remote_repository, installed_repository = \
            create_repositories(remote_entries=entries)

        # When
        with mock_print() as m:
            info_option(remote_repository, installed_repository,
                        "enstaller")
        # Then
        self.assertMultiLineEqual(m.value, r_output)


class TestUpdatesCheck(unittest.TestCase):
    def test_whats_new_no_new_epd(self):
        # Given
        r_output = textwrap.dedent("""\
            Name                 installed            available
            ============================================================
            scipy                0.12.0-1             0.13.0-1
            numpy                1.7.1-1              1.7.1-2
            """)
        installed_entries = [
            dummy_installed_package_factory("numpy", "1.7.1", 1),
            dummy_installed_package_factory("scipy", "0.12.0", 1)
        ]
        remote_entries = [
            dummy_repository_package_factory("numpy", "1.7.1", 2),
            dummy_repository_package_factory("scipy", "0.13.0", 1)
        ]

        remote, installed = create_repositories(remote_entries,
                                                installed_entries)

        # When
        with mock_print() as m:
            whats_new(remote, installed)

        # Then

        # FIXME: we splitlines and compared wo caring about order, as
        # the actual line order depends on dict ordering from
        # EggCollection.query_installed.
        assertCountEqual(self, m.value.splitlines(), r_output.splitlines())

    def test_whats_new_new_epd(self):
        # Given
        r_output = "EPD 7.3-2 is available. To update to it (with " \
                   "confirmation warning), run 'enpkg epd'.\n"
        installed_entries = [
            dummy_installed_package_factory("EPD", "7.2", 1),
        ]
        remote_entries = [
            dummy_repository_package_factory("EPD", "7.3", 2),
        ]

        remote, installed = create_repositories(remote_entries,
                                                installed_entries)

        # When
        with mock_print() as m:
            whats_new(remote, installed)

        # Then
        self.assertMultiLineEqual(m.value, r_output)

    def test_whats_new_no_updates(self):
        # Given
        r_output = "No new version of any installed package is available\n"

        installed_entries = [
            dummy_installed_package_factory("numpy", "1.7.1", 2),
            dummy_installed_package_factory("scipy", "0.13.0", 1)
        ]
        remote_entries = [
            dummy_repository_package_factory("numpy", "1.7.1", 1),
            dummy_repository_package_factory("scipy", "0.12.0", 1)
        ]

        remote, installed = create_repositories(remote_entries,
                                                installed_entries)

        # When
        with mock_print() as m:
            whats_new(remote, installed)

        # Then
        self.assertMultiLineEqual(m.value, r_output)

    def test_update_all_no_updates(self):
        r_output = "No new version of any installed package is available\n"
        config = Configuration()

        installed_entries = [
            dummy_installed_package_factory("numpy", "1.7.1", 2),
            dummy_installed_package_factory("scipy", "0.13.0", 1)
        ]
        remote_entries = [
            dummy_repository_package_factory("numpy", "1.7.1", 1),
            dummy_repository_package_factory("scipy", "0.12.0", 1)
        ]

        with mkdtemp() as d:
            enpkg = create_prefix_with_eggs(config, d,
                                            installed_entries, remote_entries)
            with mock_print() as m:
                update_all(enpkg, config, FakeOptions())
                self.assertMultiLineEqual(m.value, r_output)

    def test_update_all_no_epd_updates(self):
        r_output = textwrap.dedent("""\
        The following updates and their dependencies will be installed
        Name                 installed            available
        ============================================================
        scipy                0.13.0-1             0.13.2-1
        """)
        config = Configuration()

        installed_entries = [
            dummy_installed_package_factory("numpy", "1.7.1", 2),
            dummy_installed_package_factory("scipy", "0.13.0", 1),
            dummy_installed_package_factory("epd", "7.3", 1),
        ]
        remote_entries = [
            dummy_repository_package_factory("numpy", "1.7.1", 1),
            dummy_repository_package_factory("scipy", "0.13.2", 1),
            dummy_repository_package_factory("epd", "7.3", 1),
        ]

        with mkdtemp() as d:
            enpkg = create_prefix_with_eggs(config, d, installed_entries, remote_entries)
            with mock.patch("enstaller.cli.commands.install_req") as mocked_install_req:
                with mock_print() as m:
                    update_all(enpkg, config, FakeOptions())
                    self.assertMultiLineEqual(m.value, r_output)
                    mocked_install_req.assert_called()

    def test_update_all_epd_updates(self):
        r_output = textwrap.dedent("""\
        EPD 7.3-2 is available. To update to it (with confirmation warning), run 'enpkg epd'.
        The following updates and their dependencies will be installed
        Name                 installed            available
        ============================================================
        scipy                0.13.0-1             0.13.2-1
        """)
        config = Configuration()

        installed_entries = [
            dummy_installed_package_factory("numpy", "1.7.1", 2),
            dummy_installed_package_factory("scipy", "0.13.0", 1),
            dummy_installed_package_factory("epd", "7.3", 1),
        ]
        remote_entries = [
            dummy_repository_package_factory("numpy", "1.7.1", 1),
            dummy_repository_package_factory("scipy", "0.13.2", 1),
            dummy_repository_package_factory("epd", "7.3", 2),
        ]

        with mkdtemp() as d:
            enpkg = create_prefix_with_eggs(config, d, installed_entries, remote_entries)
            with mock.patch("enstaller.cli.commands.install_req") as mocked_install_req:
                with mock_print() as m:
                    update_all(enpkg, config, FakeOptions())
                    self.assertMultiLineEqual(m.value, r_output)
                    mocked_install_req.assert_called()


class TestInstallFromRequirements(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_install_from_requirements(self):
        # Given
        remote_entries = [
            dummy_repository_package_factory("numpy", "1.8.0", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 2),
            dummy_repository_package_factory("nose", "1.2.1", 2),
            dummy_repository_package_factory("nose", "1.3.0", 1)
        ]

        requirements_file = os.path.join(self.prefix, "requirements.txt")
        with open(requirements_file, "w") as fp:
            fp.write("numpy 1.8.0-1\nnose 1.2.1-1")

        config = Configuration()
        enpkg = create_prefix_with_eggs(config, self.prefix, [], remote_entries)
        args = FakeOptions()
        args.requirements = requirements_file

        # When
        with mock.patch("enstaller.cli.commands.install_req") as mocked_install_req:
            install_from_requirements(enpkg, config, args)

        # Then
        mocked_install_req.assert_has_calls(
            [mock.call(enpkg, config, "numpy 1.8.0-1", args),
             mock.call(enpkg, config, "nose 1.2.1-1", args)])



from __future__ import absolute_import
import argparse
import json
import sys

from jsonschema._reflect import namedAny
from jsonschema.validators import validator_for


def _namedAnyWithDefault(name):
    if "." not in name:
        name = "jsonschema." + name
    return namedAny(name)


def _json_file(path):
    with open(path) as file:
        return json.load(file)


parser = argparse.ArgumentParser(
    description="JSON Schema Validation CLI",
)
parser.add_argument(
    "-i", "--instance",
    action="append",
    dest="instances",
    type=_json_file,
    help="a path to a JSON instance to validate "
         "(may be specified multiple times)",
)
parser.add_argument(
    "-F", "--error-format",
    default="{error.instance}: {error.message}\n",
    help="the format to use for each error output message, specified in "
         "a form suitable for passing to str.format, which will be called "
         "with 'error' for each error",
)
parser.add_argument(
    "-V", "--validator",
    type=_namedAnyWithDefault,
    help="the fully qualified object name of a validator to use, or, for "
         "validators that are registered with jsonschema, simply the name "
         "of the class.",
)
parser.add_argument(
    "schema",
    help="the JSON Schema to validate with",
    type=_json_file,
)


def parse_args(args):
    arguments = vars(parser.parse_args(args=args or ["--help"]))
    if arguments["validator"] is None:
        arguments["validator"] = validator_for(arguments["schema"])
    return arguments


def main(args=sys.argv[1:]):
    sys.exit(run(arguments=parse_args(args=args)))


def run(arguments, stdout=sys.stdout, stderr=sys.stderr):
    error_format = arguments["error_format"]
    validator = arguments["validator"](schema=arguments["schema"])
    errored = False
    for instance in arguments["instances"] or ():
        for error in validator.iter_errors(instance):
            stderr.write(error_format.format(error=error))
            errored = True
    return errored



######################## BEGIN LICENSE BLOCK ########################
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA
######################### END LICENSE BLOCK #########################

__version__ = "2.2.1"
from sys import version_info


def detect(aBuf):
    if ((version_info < (3, 0) and isinstance(aBuf, unicode)) or
            (version_info >= (3, 0) and not isinstance(aBuf, bytes))):
        raise ValueError('Expected a bytes object, not a unicode object')

    from . import universaldetector
    u = universaldetector.UniversalDetector()
    u.reset()
    u.feed(aBuf)
    u.close()
    return u.result



#
# (C) Copyright 2014 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is open source software distributed according to the terms in
# LICENSE.txt
#

"""
A pure python, ctypes-based replacement for win32cred features required by
keyring.
"""
from __future__ import absolute_import

import ctypes
from ctypes import POINTER, Structure, c_void_p, c_wchar_p
from ctypes.wintypes import (
    BOOL, DWORD, FILETIME, LPCWSTR)

from ._common import LPBYTE
from ._util import function_factory, check_zero_factory


class CREDENTIAL(Structure):
    _fields_ = [
        ("Flags", DWORD),
        ("Type", DWORD),
        ("TargetName", c_wchar_p),
        ("Comment", c_wchar_p),
        ("LastWritten", FILETIME),
        ("CredentialBlobSize", DWORD),
        ("CredentialBlob", LPBYTE),
        ("Persist", DWORD),
        ("_DO_NOT_USE_AttributeCount", DWORD),
        ("__DO_NOT_USE_Attribute", c_void_p),
        ("TargetAlias", c_wchar_p),
        ("UserName", c_wchar_p)]
PCREDENTIAL = POINTER(CREDENTIAL)

SUPPORTED_CREDKEYS = set((
    'Type', 'TargetName', 'Persist', 'UserName', 'Comment', 'CredentialBlob'))

advapi = ctypes.windll.advapi32

_CredWrite = function_factory(
    advapi.CredWriteW,
    [PCREDENTIAL, DWORD],
    BOOL,
    check_zero_factory("CredWrite"))

_CredRead = function_factory(
    advapi.CredReadW,
    [LPCWSTR, DWORD, DWORD, POINTER(PCREDENTIAL)],
    BOOL,
    check_zero_factory("CredRead"))

_CredDelete = function_factory(
    advapi.CredDeleteW,
    [LPCWSTR, DWORD, DWORD],
    BOOL,
    check_zero_factory("CredDelete"))



#
# (C) Copyright 2014 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is open source software distributed according to the terms in
# LICENSE.txt
#

""" Common constants and wrapped functions
"""
from __future__ import absolute_import

import ctypes
from ctypes import pythonapi, POINTER, c_void_p, py_object
from ctypes.wintypes import BYTE, UINT

from .compat import PY3
from ._util import function_factory

PPy_UNICODE = c_void_p


kernel32 = ctypes.windll.kernel32


if ctypes.sizeof(ctypes.c_long) == ctypes.sizeof(ctypes.c_void_p):
    LONG_PTR = ctypes.c_long
elif ctypes.sizeof(ctypes.c_longlong) == ctypes.sizeof(ctypes.c_void_p):
    LONG_PTR = ctypes.c_longlong

LPBYTE = POINTER(BYTE)

if PY3:
    _PyBytes_FromStringAndSize = function_factory(
        pythonapi.PyBytes_FromStringAndSize,
        return_type=py_object)
else:
    _PyBytes_FromStringAndSize = function_factory(
        pythonapi.PyString_FromStringAndSize,
        return_type=py_object)

_GetACP = function_factory(
    kernel32.GetACP,
    None,
    UINT)



from __future__ import absolute_import

import abc
import base64
import re
import textwrap

from egginst._compat import with_metaclass
from enstaller.errors import InvalidConfiguration
from enstaller.vendor import requests

from .auth_managers import BroodBearerTokenAuth


_CLEARTEXT_AUTH_R = re.compile(r'^(EPD_auth|EPD_username)\s*=.*$', re.M)
_API_TOKEN_AUTH_R = re.compile(r'^api_token\s*=.*$', re.M)


def subscription_message(config, user):
    """
    Return a 'subscription level' message based on the `user`
    information.

    Parameters
    ----------
    config : Configuration
    user : UserInfo

    Returns
    -------
    message : str
        The subscription message.
    """
    message = ""

    if user.is_authenticated:
        login = config.auth.logged_message
        subscription = "Subscription level: %s" % user.subscription_level
        name = user.first_name + ' ' + user.last_name
        name = name.strip()
        if name:
            name = ' (' + name + ')'
        message = login + name + '.\n' + subscription
    else:
        message = "You are not logged in.  To log in, type 'enpkg --userpass'."

    return message


class IAuth(with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def change_auth(self, filename):
        """
        Change the auth information in the configuration file.
        """

    @abc.abstractmethod
    def to_config_dict(self):
        """
        Return a dict that can be used for json/yaml serialization.
        """

    @abc.abstractproperty
    def cant_login_message(self):
        """
        Return a string to show when cannot authenticate.
        """

    @abc.abstractproperty
    def config_string(self):
        """
        The text to write in the configuration file for this particular
        auth.
        """

    @abc.abstractproperty
    def logged_message(self):
        """
        The string to print to indicate a user is logged in.
        """

    @abc.abstractproperty
    def request_adapter(self):
        """ Return an object that can be passed to the auth argument of
        requests Session and feunctions.
        """


class UserPasswordAuth(IAuth):
    """ Simple clear text username/password authentication."""
    @classmethod
    def from_encoded_auth(cls, encoded_auth):
        parts = base64.decodestring(encoded_auth.encode("utf8")). \
            decode("utf8"). \
            split(":")
        if len(parts) == 2:
            return cls(*parts)
        else:
            raise InvalidConfiguration("Invalid auth line")

    def __init__(self, username, password):
        if username is None or username == "":
            msg = "Invalid username: {0!r}".format(username)
            raise InvalidConfiguration(msg)

        self.username = username
        self.password = password

    def change_auth(self, filename):
        with open(filename, 'r') as fi:
            data = fi.read()

        authline = 'EPD_auth = \'%s\'' % self._encoded_auth

        if _API_TOKEN_AUTH_R.search(data):
            data = _API_TOKEN_AUTH_R.sub("", data)

        if _CLEARTEXT_AUTH_R.search(data):
            data = _CLEARTEXT_AUTH_R.sub(authline, data)
        else:
            lines = data.splitlines()
            lines.append(authline)
            data = '\n'.join(lines) + '\n'

        with open(filename, 'w') as fo:
            fo.write(data)

    def to_config_dict(self):
        return {"kind": "simple",
                "username": self.username,
                "password": self.password}

    @property
    def config_string(self):
        authline = "EPD_auth = '%s'" % self._encoded_auth
        auth_section = textwrap.dedent("""
        # A Canopy / EPD subscriber authentication is required to access the
        # Canopy / EPD repository.  To change your credentials, use the 'enpkg
        # --userpass' command, which will ask you for your email address
        # password.
        %s
        """ % authline)
        return auth_section

    @property
    def cant_login_message(self):
        return "Could not authenticate as '{0}'".format(self.username)

    @property
    def logged_message(self):
        return "You are logged in as '{0}'".format(self.username)

    @property
    def request_adapter(self):
        return requests.auth.HTTPBasicAuth(self.username, self.password)

    # ------------------
    # Private properties
    # ------------------
    @property
    def _encoded_auth(self):
        """
        Auth information, encoded as expected by EPD_auth.
        """
        return _encode_auth(self.username, self.password)

    def __eq__(self, other):
        return self.__class__ == other.__class__ \
            and self.username == other.username \
            and self.password == other.password


class APITokenAuth(IAuth):
    def __init__(self, api_token):
        self._api_token = api_token

    def change_auth(self, filename):
        with open(filename, "rt") as fp:
            data = fp.read()

        if _CLEARTEXT_AUTH_R.search(data):
            data = _CLEARTEXT_AUTH_R.sub("", data)

        m = _API_TOKEN_AUTH_R.search(data)
        if m:
            data = _API_TOKEN_AUTH_R.sub(self.config_string, data)
        else:
            lines = data.splitlines()
            lines.append(self.config_string)
            data = "\n".join(lines)

        with open(filename, "wt") as fp:
            fp.write(data)

    def to_config_dict(self):
        return {"kind": "token", "api_token": self.api_token}

    @property
    def cant_login_message(self):
        msg = ("Could not authenticate with the given token: check your token"
               " settings")
        return msg

    @property
    def config_string(self):
        return "api_token = '{0}'".format(self._api_token)

    @property
    def logged_message(self):
        return "logged in using API token"

    @property
    def request_adapter(self):
        return BroodBearerTokenAuth(self._api_token)

    def __eq__(self, other):
        return self.__class__ == other.__class__ \
            and self._api_token == other._api_token


def _encode_string_base64(s):
    return base64.encodestring(s.encode("utf8")).decode("utf8")


def _encode_auth(username, password):
    s = "{0}:{1}".format(username, password)
    return _encode_string_base64(s).rstrip()



"""
core.py

Created by Kang Zhang on 2009-07-09
"""
import os
import sys
import warnings
import logging

from .py27compat import configparser

from . import logger
from . import backend
from .util import platform_ as platform

log = logging.getLogger(__name__)

_keyring_backend = None

def set_keyring(keyring):
    """Set current keyring backend.
    """
    global _keyring_backend
    if not isinstance(keyring, backend.KeyringBackend):
        raise TypeError("The keyring must be a subclass of KeyringBackend")
    _keyring_backend = keyring


def get_keyring():
    """Get current keyring backend.
    """
    return _keyring_backend


def get_password(service_name, username):
    """Get password from the specified service.
    """
    return _keyring_backend.get_password(service_name, username)


def set_password(service_name, username, password):
    """Set password for the user in the specified service.
    """
    _keyring_backend.set_password(service_name, username, password)


def delete_password(service_name, username):
    """Delete the password for the user in the specified service.
    """
    _keyring_backend.delete_password(service_name, username)


def init_backend():
    """
    Load a keyring specified in the config file or infer the best available.
    """
    _load_library_extensions()
    set_keyring(load_config() or _get_best_keyring())


def _get_best_keyring():
    """
    Return the best keyring backend for the given environment based on
    priority.
    """
    keyrings = backend.get_all_keyring()
    # rank by priority
    keyrings.sort(key = lambda x: -x.priority)
    # get the most recommended one
    return keyrings[0]


def load_keyring(keyring_path, keyring_name):
    """
    Load the specified keyring by name (a fully-qualified name to the
    keyring, such as 'keyring.backends.file.PlaintextKeyring')

    `keyring_path` is an additional, optional search path and may be None.
    **deprecated** In the future, keyring_path must be None.
    """
    module_name, sep, class_name = keyring_name.rpartition('.')
    if keyring_path is not None and keyring_path not in sys.path:
        warnings.warn("keyring_path is deprecated and should always be None",
            DeprecationWarning)
        sys.path.insert(0, keyring_path)
    __import__(module_name)
    module = sys.modules[module_name]
    class_ = getattr(module, class_name)
    # invoke the priority to ensure it is viable, or raise a RuntimeError
    class_.priority
    return class_()


def load_config():
    """Load a keyring using the config file.

    The config file can be in the current working directory, or in the user's
    home directory.
    """
    keyring = None

    filename = 'keyringrc.cfg'

    local_path = os.path.join(os.getcwd(), filename)
    config_path = os.path.join(platform.config_root(), filename)

    # search from current working directory and the data root
    keyring_cfg_candidates = [local_path, config_path]

    # initialize the keyring_config with the first detected config file
    keyring_cfg = None
    for path in keyring_cfg_candidates:
        keyring_cfg = path
        if os.path.exists(path):
            break

    if os.path.exists(keyring_cfg):
        config = configparser.RawConfigParser()
        config.read(keyring_cfg)
        _load_keyring_path(config)

        # load the keyring class name, and then load this keyring
        try:
            if config.has_section("backend"):
                keyring_name = config.get("backend", "default-keyring").strip()
            else:
                raise configparser.NoOptionError('backend', 'default-keyring')

            keyring = load_keyring(None, keyring_name)
        except (configparser.NoOptionError, ImportError):
            logger.warning("Keyring config file contains incorrect values.\n" +
                           "Config file: %s" % keyring_cfg)

    return keyring

def _load_keyring_path(config):
    "load the keyring-path option (if present)"
    try:
        path = config.get("backend", "keyring-path").strip()
        sys.path.insert(0, path)
    except (configparser.NoOptionError, configparser.NoSectionError):
        pass

def _load_library_extensions():
    """
    Locate all setuptools entry points by the name 'keyring backends'
    and initialize them.

    Any third-party library may register an entry point by adding the
    following to their setup.py::

        entry_points = {
            'keyring backends': [
                'plugin name = mylib.mymodule:initialize_func',
            ],
        },

    `plugin name` can be anything.
    `initialize_func` is optional and will be invoked by keyring on startup.

    Most plugins will simply provide or import a KeyringBackend in `mymodule`.
    """
    group = 'keyring backends'
    try:
        pkg_resources = __import__('pkg_resources')
    except ImportError:
        return
    entry_points = pkg_resources.iter_entry_points(group=group)
    for ep in entry_points:
        try:
            log.info('Loading keyring backends from %s', ep.name)
            init_func = ep.load()
            if callable(init_func):
                init_func()
        except Exception as exc:
            log.exception("Error initializing plugin %s (%s).", ep, exc)

# init the _keyring_backend
init_backend()



from __future__ import absolute_import

import os.path
import shutil
import sys
import tempfile
import textwrap

if sys.version_info < (2, 7):
    import unittest2 as unittest
    # FIXME: this looks quite fishy. On 2.6, with unittest2, the assertRaises
    # context manager does not contain the actual exception object ?
    def exception_code(ctx):
        return ctx.exception
else:
    import unittest
    def exception_code(ctx):
        return ctx.exception.code

import mock

from enstaller.main import main
from enstaller.tests.common import mock_print

from .common import (fake_empty_resolve, fake_configuration_and_auth,
                     enstaller_version, authenticated_config,
                     raw_input_always_yes, remote_enstaller_available)

class TestEnstallerMainActions(unittest.TestCase):
    def setUp(self):
        self.d = tempfile.mkdtemp()
        self.config = os.path.join(self.d, ".enstaller4rc")

    def tearDown(self):
        shutil.rmtree(self.d)

    @authenticated_config
    @raw_input_always_yes
    @enstaller_version("4.6.1")
    @remote_enstaller_available(["4.6.2"])
    def test_automatic_update(self):
        r_output = textwrap.dedent("""\
            Enstaller has been updated.
            Please re-run your previous command.
        """)

        with mock_print() as m:
            with mock.patch("enstaller.main.update_enstaller"):
                with mock.patch("enstaller.main.install_req"):
                    main([""])
        self.assertMultiLineEqual(m.value, r_output)

    @authenticated_config
    @raw_input_always_yes
    @enstaller_version("4.6.1")
    @remote_enstaller_available(["4.6.2"])
    def test_enstaller_in_req(self):
        r_output = textwrap.dedent("""\
            Enstaller has been updated.
            Please re-run your previous command.
        """)

        with mock_print() as m:
            with mock.patch("enstaller.main.inplace_update"):
                main(["enstaller"])
        self.assertMultiLineEqual(m.value, r_output)

    @authenticated_config
    @raw_input_always_yes
    @enstaller_version("4.6.3")
    @remote_enstaller_available(["4.6.2"])
    @mock.patch("enstaller.main.logger")
    def test_updated_enstaller(self, logger):
        with mock.patch("enstaller.main.install_req"):
            main([""])
        logger.info.assert_called_with('prefix: %r',
                                       os.path.normpath(sys.prefix))

    @authenticated_config
    @raw_input_always_yes
    @enstaller_version("4.6.3")
    @remote_enstaller_available(["4.6.2"])
    def test_updated_enstaller_in_req(self):
        with mock_print() as m:
            with mock.patch("enstaller.main.install_req"):
                main(["enstaller"])
        self.assertMultiLineEqual(m.value, "")


class TestEnstallerInstallActions(unittest.TestCase):
    @fake_configuration_and_auth
    @fake_empty_resolve
    def test_install_numpy(self):
        main(["numpy"])

    @fake_configuration_and_auth
    @fake_empty_resolve
    def test_install_epd(self):
        with mock.patch("enstaller.main.epd_install_confirm") as m:
            main(["epd"])
        self.assertTrue(m.called)

    @fake_configuration_and_auth
    def test_remove_epd_fails(self):
        with mock.patch("enstaller.main.epd_install_confirm"):
            with mock.patch("enstaller.main.install_req"):
                with self.assertRaises(SystemExit) as e:
                    main(["--remove", "epd"])
                    self.assertNotEqual(exception_code(e), 0)

    @fake_configuration_and_auth
    def test_install_epd_and_other(self):
        with mock.patch("enstaller.main.epd_install_confirm"):
            with mock.patch("enstaller.main.install_req"):
                with self.assertRaises(SystemExit) as e:
                    main(["epd", "numpy"])
                self.assertNotEqual(exception_code(e), 0)

    @fake_configuration_and_auth
    def test_remove(self):
        with mock.patch("enstaller.main.Enpkg.execute"):
            main(["--remove", "numpy"])



from __future__ import absolute_import

import os

from os.path import abspath, dirname, join

from enstaller.repository import Repository, RepositoryPackageMetadata
from enstaller.utils import PY_VER
from egginst.vendor.six.moves import unittest

from enstaller.indexed_repo.metadata import parse_depend_index

from ..resolve import Resolve
from ..requirement import Requirement


INDEX_REPO_DIR = abspath(join(dirname(__file__), os.pardir, os.pardir,
                              "indexed_repo", "tests"))


def _old_style_index_to_packages(index_path, python_version):
    with open(index_path) as fp:
        index_data = fp.read()

    packages = []
    index = parse_depend_index(index_data)
    for key, spec in index.items():
        spec['name'] = spec['name'].lower()
        spec['type'] = 'egg'
        spec['store_location'] = os.path.dirname(index_path)

        if python_version == "*":
            package = RepositoryPackageMetadata.from_json_dict(key, spec)
            packages.append(package)
        elif spec["python"] in (None, python_version):
            package = RepositoryPackageMetadata.from_json_dict(key, spec)
            packages.append(package)

    return packages


def _old_style_indices_to_repository(indices, python_version=PY_VER):
    repository = Repository()
    for index in indices:
        for package in _old_style_index_to_packages(index, python_version):
            repository.add_package(package)
    return repository


def eggs_rs(c, req_string):
    return c.install_sequence(Requirement(req_string))


class TestChain0(unittest.TestCase):
    def _resolve_factory(self, python_version):
        indices = [join(INDEX_REPO_DIR, fn) for fn in ['index-add.txt',
                                                       'index-5.1.txt',
                                                       'index-5.0.txt',
                                                       'index-cycle.txt']]
        repo = _old_style_indices_to_repository(indices, python_version)
        return Resolve(repo)

    def test_25(self):
        resolve = self._resolve_factory("2.5")
        self.assertEqual(eggs_rs(resolve, 'SciPy 0.8.0.dev5698'),
                         ['freetype-2.3.7-1.egg', 'libjpeg-7.0-1.egg',
                          'numpy-1.3.0-1.egg', 'PIL-1.1.6-4.egg',
                          'scipy-0.8.0.dev5698-1.egg'])

        self.assertEqual(eggs_rs(resolve, 'SciPy'),
                         ['numpy-1.3.0-1.egg', 'scipy-0.8.0-1.egg'])

        self.assertEqual(eggs_rs(resolve, 'epdcore'),
                         ['AppInst-2.0.4-1.egg', 'numpy-1.3.0-1.egg',
                          'scipy-0.8.0-1.egg', 'EPDCore-1.2.5-1.egg'])

    def test_26(self):
        resolve = self._resolve_factory("2.6")
        self.assertEqual(eggs_rs(resolve, 'SciPy'),
                         ['numpy-1.3.0-2.egg', 'scipy-0.8.0-2.egg'])

        self.assertEqual(eggs_rs(resolve, 'epdcore'),
                         ['numpy-1.3.0-2.egg', 'scipy-0.8.0-2.egg',
                          'EPDCore-2.0.0-1.egg'])


class TestChain1(unittest.TestCase):
    def _resolve_factory(self, python_version=PY_VER):
        indices = [join(INDEX_REPO_DIR, name, 'index-7.1.txt') for name
                   in ('epd', 'gpl')]
        repo = _old_style_indices_to_repository(indices, python_version)
        return Resolve(repo)

    def test_get_repo(self):
        resolve = self._resolve_factory()
        for req_string, repo_name in [
            ('MySQL_python', 'gpl'),
            ('bitarray', 'epd'),
            ('foobar', None),
        ]:
            resolve._latest_egg(Requirement(req_string))

    def test_get_dist(self):
        resolve = self._resolve_factory("2.7")
        for req_string, repo_name, egg in [
            ('MySQL_python',  'gpl', 'MySQL_python-1.2.3-2.egg'),
            ('numpy',         'epd', 'numpy-1.6.0-3.egg'),
            ('swig',          'epd', 'swig-1.3.40-2.egg'),
            ('swig 1.3.36',   'epd', 'swig-1.3.36-3.egg'),
            ('swig 1.3.40-1', 'epd', 'swig-1.3.40-1.egg'),
            ('swig 1.3.40-2', 'epd', 'swig-1.3.40-2.egg'),
            ('foobar', None, None),
        ]:
            self.assertEqual(resolve._latest_egg(Requirement(req_string)), egg)

    def test_reqs_dist(self):
        resolve = self._resolve_factory("*")
        self.assertEqual(resolve._dependencies_from_egg('FiPy-2.1-1.egg'),
                         set([Requirement('distribute'),
                              Requirement('scipy'),
                              Requirement('numpy'),
                              Requirement('pysparse 1.2.dev203')]))

    def test_root(self):
        resolve = self._resolve_factory("2.7")
        self.assertEqual(resolve.install_sequence(Requirement('numpy 1.5.1'),
                                                  mode='root'),
                         ['numpy-1.5.1-2.egg'])

        self.assertEqual(resolve.install_sequence(Requirement('numpy 1.5.1-1'),
                                                  mode='root'),
                         ['numpy-1.5.1-1.egg'])

    def test_order1(self):
        resolve = self._resolve_factory("2.7")
        self.assertEqual(resolve.install_sequence(Requirement('numpy')),
                         ['MKL-10.3-1.egg', 'numpy-1.6.0-3.egg'])

    def test_order2(self):
        resolve = self._resolve_factory("2.7")
        self.assertEqual(resolve.install_sequence(Requirement('scipy')),
                         ['MKL-10.3-1.egg', 'numpy-1.5.1-2.egg',
                          'scipy-0.9.0-1.egg'])


class TestChain2(unittest.TestCase):
    def _resolve_factory(self, python_version=PY_VER):
        indices = [join(INDEX_REPO_DIR, name, 'index-7.1.txt') for name
                   in ('open', 'runner', 'epd')]
        self.repo = _old_style_indices_to_repository(indices, python_version)

        return Resolve(self.repo)

    def test_flat_recur1(self):
        resolve = self._resolve_factory("2.7")
        d1 = resolve.install_sequence(Requirement('openepd'), mode='flat')
        d2 = resolve.install_sequence(Requirement('openepd'), mode='recur')
        self.assertEqual(d1, d2)
        d3 = resolve.install_sequence(Requirement('foo'), mode='recur')
        self.assertEqual(d2[:-1], d3[:-1])

    def test_flat_recur2(self):
        resolve = self._resolve_factory("2.7")
        for rs in 'epd 7.0', 'epd 7.0-1', 'epd 7.0-2':
            d1 = resolve.install_sequence(Requirement(rs), mode='flat')
            d2 = resolve.install_sequence(Requirement(rs), mode='recur')
            self.assertEqual(d1, d2)

    def test_multiple_reqs(self):
        resolve = self._resolve_factory("2.7")
        lst = resolve.install_sequence(Requirement('ets'))
        self.assert_('numpy-1.5.1-2.egg' in lst)


class TestCycle(unittest.TestCase):
    """Avoid an infinite recursion when the dependencies contain a cycle."""

    def _resolve_factory(self, python_version):
        indices = [join(INDEX_REPO_DIR, 'index-cycle.txt')]
        repo = _old_style_indices_to_repository(indices, python_version)
        return Resolve(repo)

    def test_cycle(self):
        resolve = self._resolve_factory("2.5")
        try:
            eg = eggs_rs(resolve, 'cycleParent 2.0-5')
        except Exception as e:
            self.assertIn("Loop", str(e),
                          "unexpected exception message " + repr(e))
        else:
            self.assertIsNone(eg,
                              "dependency cycle did not trigger an exception "
                              + repr(eg))



import contextlib
import functools
import os
import tempfile

import mock

from egginst._compat import PY2

from enstaller.config import Configuration
from enstaller.repository import Repository
from enstaller.plat import custom_plat
from enstaller.tests.common import (authenticated_config,
                                    dummy_repository_package_factory)
from enstaller.vendor import responses


def _dont_write_default_configuration(f):
    return mock.patch("enstaller.main.write_default_config",
                      lambda filename, use_keyring=None: None)(f)


def without_any_configuration(f):
    """
    When this decorator is applied, enstaller.main will behave as if no default
    configuration is found anywhere, and no default configuration will be
    written in $HOME.
    """
    @functools.wraps(f)
    def wrapper(*a, **kw):
        with tempfile.NamedTemporaryFile(delete=False) as fp:
            pass
        try:
            with mock.patch("enstaller.main._ensure_config_path",
                            return_value=fp.name):
                return f(*a, **kw)
        finally:
            os.unlink(fp.name)
    return wrapper


def mock_install_req(f):
    """
    Decorating a function/class with this decorator will mock install_req
    completely within enstaller.main function.
    """
    return mock.patch("enstaller.main.install_req", mock.Mock())(f)


def fake_empty_resolve(f):
    @functools.wraps(f)
    def wrapper(*a, **kw):
        with mock.patch("enstaller.enpkg.Solver.resolve", return_value=[]):
            return f(*a, **kw)
    return wrapper


def empty_index(f):
    @functools.wraps(f)
    def wrapper(*a, **kw):
        with mock.patch("enstaller.main.repository_factory",
                        return_value=Repository()):
            return f(*a, **kw)
    return wrapper


@contextlib.contextmanager
def set_env_vars(**kw):
    old_env = os.environ.copy()
    try:
        yield os.environ.update(**kw)
    finally:
        os.environ = old_env


@contextlib.contextmanager
def use_given_config_context(filename):
    """
    When this decorator is applied, enstaller.main will use the given filename
    as its configuration file.
    """
    with mock.patch("enstaller.main._ensure_config_path",
                    return_value=filename) as context:
        yield context


def fake_configuration_and_auth(f):
    config = Configuration()
    config.update(auth=("john", "doe"))
    @functools.wraps(f)
    @responses.activate
    def wrapper(*a, **kw):
        responses.add(responses.GET,
                      "https://api.enthought.com/eggs/{0}/index.json". \
                        format(custom_plat),
                      body="{}", status=200,
                      content_type='application/json')

        # FIXME: we create a dummy store to bypass store authentication in
        # Enpkg ctor. Will be fixed once Enpkg, repository, stores are clearly
        # separated.
        with mock.patch("enstaller.main.Configuration.from_file",
                        return_value=config):
            with mock.patch("enstaller.main.ensure_authenticated_config",
                            return_value=True):
                return without_any_configuration(f)(*a, **kw)
    return wrapper


def enstaller_version(version, is_released=True):
    wrap1 = mock.patch("enstaller.__version__", version)
    wrap2 = mock.patch("enstaller.__is_released__", is_released)
    def dec(f):
        return wrap1(wrap2(f))
    return dec

def remote_enstaller_available(versions):
    repository = Repository()
    for version in versions:
        package = dummy_repository_package_factory("enstaller", version, build=1)
        repository.add_package(package)

    def dec(f):
        @functools.wraps(f)
        def wrapper(*a, **kw):
            with mock.patch("enstaller.main.repository_factory",
                            return_value=repository):
                return f(*a, **kw)
        return wrapper
    return dec

def raw_input_always_yes(f):
    if PY2:
        wrap = mock.patch("__builtin__.raw_input", lambda ignored: "y")
    else:
        wrap = mock.patch("builtins.input", lambda ignored: "y")
    return wrap(f)




from ctypes import Structure, POINTER, c_void_p, cast, create_string_buffer, \
    c_char_p, byref, memmove
from ctypes import windll, WinDLL, WINFUNCTYPE
try:
    from ctypes import wintypes
except ValueError:
    # see http://bugs.python.org/issue16396
    raise ImportError("wintypes")

from ..util.escape import u

# Crypto API ctypes bindings

class DATA_BLOB(Structure):
    _fields_ = [('cbData', wintypes.DWORD),
                ('pbData', POINTER(wintypes.BYTE))]


class CRYPTPROTECT_PROMPTSTRUCT(Structure):
    _fields_ = [('cbSize', wintypes.DWORD),
                ('dwPromptFlags', wintypes.DWORD),
                ('hwndApp', wintypes.HWND),
                ('szPrompt', POINTER(wintypes.WCHAR))]

# Flags for CRYPTPROTECT_PROMPTSTRUCT

CRYPTPROTECT_PROMPT_ON_UNPROTECT = 1
CRYPTPROTECT_PROMPT_ON_PROTECT = 2

# Flags for CryptProtectData/CryptUnprotectData

CRYPTPROTECT_UI_FORBIDDEN = 0x01
CRYPTPROTECT_LOCAL_MACHINE = 0x04
CRYPTPROTECT_CRED_SYNC = 0x08
CRYPTPROTECT_AUDIT = 0x10
CRYPTPROTECT_NO_RECOVERY = 0x20
CRYPTPROTECT_VERIFY_PROTECTION = 0x40
CRYPTPROTECT_CRED_REGENERATE = 0x80

# Crypto API Functions

_dll = WinDLL('CRYPT32.DLL')

CryptProtectData = WINFUNCTYPE(wintypes.BOOL,
                               POINTER(DATA_BLOB),
                               POINTER(wintypes.WCHAR),
                               POINTER(DATA_BLOB),
                               c_void_p,
                               POINTER(CRYPTPROTECT_PROMPTSTRUCT),
                               wintypes.DWORD,
                               POINTER(DATA_BLOB))(('CryptProtectData', _dll))

CryptUnprotectData = WINFUNCTYPE(wintypes.BOOL,
                                 POINTER(DATA_BLOB),
                                 POINTER(wintypes.WCHAR),
                                 POINTER(DATA_BLOB),
                                 c_void_p,
                                 POINTER(CRYPTPROTECT_PROMPTSTRUCT),
                                 wintypes.DWORD, POINTER(DATA_BLOB))(
                                         ('CryptUnprotectData', _dll))

# Functions


def encrypt(data, non_interactive=0):
    blobin = DATA_BLOB(cbData=len(data),
                       pbData=cast(c_char_p(data),
                                   POINTER(wintypes.BYTE)))
    blobout = DATA_BLOB()

    if not CryptProtectData(byref(blobin),
                            u('python-keyring-lib.win32crypto'),
                            None, None, None,
                            CRYPTPROTECT_UI_FORBIDDEN,
                            byref(blobout)):
        raise OSError("Can't encrypt")

    encrypted = create_string_buffer(blobout.cbData)
    memmove(encrypted, blobout.pbData, blobout.cbData)
    windll.kernel32.LocalFree(blobout.pbData)
    return encrypted.raw


def decrypt(encrypted, non_interactive=0):
    blobin = DATA_BLOB(cbData=len(encrypted),
                       pbData=cast(c_char_p(encrypted),
                                   POINTER(wintypes.BYTE)))
    blobout = DATA_BLOB()

    if not CryptUnprotectData(byref(blobin),
                              u('python-keyring-lib.win32crypto'),
                              None, None, None,
                              CRYPTPROTECT_UI_FORBIDDEN,
                              byref(blobout)):
        raise OSError("Can't decrypt")

    data = create_string_buffer(blobout.cbData)
    memmove(data, blobout.pbData, blobout.cbData)
    windll.kernel32.LocalFree(blobout.pbData)
    return data.raw



import sys

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

import mock

from enstaller.main import main_noexc

from .common import fake_configuration_and_auth


class TestRevert(unittest.TestCase):
    @fake_configuration_and_auth
    def test_simple(self):
        with mock.patch("enstaller.main.Enpkg.revert_actions"):
            with self.assertRaises(SystemExit) as e:
                main_noexc(["--revert", "10"])
            self.assertEqual(e.exception.code, 0)

    @fake_configuration_and_auth
    def test_no_actions(self):
        with mock.patch("enstaller.main.Enpkg") as enpkg:
            enpkg.return_value.configure_mock(**{"revert_actions.return_value": []})
            with self.assertRaises(SystemExit) as e:
                main_noexc(["--revert", "10"])
            self.assertEqual(e.exception.code, 0)



from .pep386_workaround import PEP386WorkaroundVersion


class EnpkgVersion(object):
    @classmethod
    def from_upstream_and_build(cls, upstream, build):
        """ Creates a new EnpkgVersion from the upstream string and the
        build number.

        Parameters
        ----------
        upstream : str
            The upstream version (e.g. '1.3.0')
        build : int
            The build number
        """
        upstream = PEP386WorkaroundVersion.from_string(upstream)
        return cls(upstream, build)

    @classmethod
    def from_string(cls, version_string):
        """ Creates a new EnpkgVersion from its string representation.

        Parameters
        ----------
        version_string : str
            the version string.
        """
        parts = version_string.rsplit("-")
        if len(parts) < 2:
            msg = "Invalid version format: {0!r}".format(version_string)
            raise ValueError(msg)

        try:
            build = int(parts[1])
        except ValueError:
            raise ValueError("Invalid build number: {0!r}".format(parts[1]))

        return cls.from_upstream_and_build(parts[0], build)

    def __init__(self, upstream, build):
        """ Creates a new EnpkgVersion instance

        Parameters
        ----------
        upstream : PEP386WorkaroundVersion
            The upstream version
        build : int
            The build number
        """
        self.upstream = upstream
        self.build = build

        self._parts = upstream, build

    def __str__(self):
        return str(self.upstream) + "-" + str(self.build)

    def __hash__(self):
        return hash(self._parts)

    def _cannot_compare(self, other):
        msg = "Cannot compare {0!r} and {1!r}"
        raise TypeError(msg.format(type(self), type(other)))

    def __eq__(self, other):
        if not isinstance(other, EnpkgVersion):
            self._cannot_compare(other)
        return self._parts == other._parts

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, EnpkgVersion):
            self._cannot_compare(other)
        return self._parts < other._parts

    def __le__(self, other):
        return not self.__gt__(other)

    def __gt__(self, other):
        if not isinstance(other, EnpkgVersion):
            self._cannot_compare(other)
        return self._parts > other._parts

    def __ge__(self, other):
        return not self.__lt__(other)



from egginst.main import EggInst
from egginst.tests.common import DUMMY_EGG, create_venv, mkdtemp
from egginst.vendor.six.moves import unittest

from enstaller.freeze import get_freeze_list


class TestFreeze(unittest.TestCase):
    def test_simple(self):
        # Given
        path = DUMMY_EGG
        with mkdtemp() as tempdir:
            create_venv(tempdir)
            installer = EggInst(path, prefix=tempdir)
            installer.install()

            # When
            freeze_list = get_freeze_list([tempdir])

        self.assertEqual(freeze_list, ["dummy 1.0.1-1"])



import os
import shutil

from egginst._compat import assertCountEqual
from egginst.main import EggInst, setuptools_egg_info_dir
from egginst.repair_broken_egg_info import EggInfoDirFixer, repair
from egginst.testing_utils import slow
from egginst.tests.common import (DUMMY_EGG, DUMMY_EGG_WITH_ENTRY_POINTS,
                                  DUMMY_EGG_WITH_APPINST)
from egginst.tests.common import tempfile
from egginst.utils import compute_md5
from egginst.vendor.six.moves import unittest


class TestEggInfoDirFixer(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def _install_egg(self, egg_path):
        installer = EggInst(egg_path, prefix=self.prefix)
        installer.install()

    def _install_egg_empty_egg_info_dir(self, egg_path):
        installer = EggInst(egg_path, prefix=self.prefix)
        installer.install()

        egg_info_dir = os.path.join(installer.site_packages,
                                    setuptools_egg_info_dir(egg_path))
        shutil.rmtree(egg_info_dir)
        os.makedirs(egg_info_dir)

    def _install_egg_file_egg_info_dir(self, egg_path):
        installer = EggInst(egg_path, prefix=self.prefix)
        installer.install()

        egg_info_dir = os.path.join(installer.site_packages,
                                    setuptools_egg_info_dir(egg_path))
        shutil.rmtree(egg_info_dir)
        with open(egg_info_dir, "w") as fp:
            fp.write("")

    def _egg_info_path(self, egg):
        fixer = EggInfoDirFixer(egg, prefix=self.prefix)
        return fixer.egg_info_dir

    @slow
    def test_needs_repair_non_broken(self):
        """
        Ensure we detect correctly non-broken .egg-info directories
        """
        # Given
        egg_path = DUMMY_EGG
        self._install_egg(egg_path)

        # When
        fixer = EggInfoDirFixer(egg_path, prefix=self.prefix)

        # Then
        self.assertFalse(fixer.needs_repair())

    @slow
    def test_needs_repair_empty_dir(self):
        """
        Ensure we detect correctly broken .egg-info directories as installed in
        enstaller == 4.6.3 on windows.
        """
        # Given
        egg_path = DUMMY_EGG
        self._install_egg_empty_egg_info_dir(egg_path)

        # When
        fixer = EggInfoDirFixer(egg_path, prefix=self.prefix)

        # Then
        self.assertTrue(fixer.needs_repair())

    @slow
    def test_needs_repair_file_hack(self):
        """
        Ensure we detect correctly broken .egg-info files as installed in
        enstaller < 4.6.3.
        """
        # Given
        egg_path = DUMMY_EGG
        self._install_egg_file_egg_info_dir(egg_path)

        # When
        fixer = EggInfoDirFixer(egg_path, prefix=self.prefix)

        # Then
        self.assertTrue(fixer.needs_repair())

    @slow
    def test_repair_empty_dir(self):
        # Given
        egg_path = DUMMY_EGG
        self._install_egg_empty_egg_info_dir(egg_path)

        # When
        fixer = EggInfoDirFixer(egg_path, prefix=self.prefix)
        fixer.repair()

        # Then
        assertCountEqual(self, os.listdir(fixer.egg_info_dir),
                         ["PKG-INFO", "egginst.json", "_info.json"])

    @slow
    def test_repair_file(self):
        # Given
        egg_path = DUMMY_EGG
        self._install_egg_file_egg_info_dir(egg_path)

        # When
        fixer = EggInfoDirFixer(egg_path, prefix=self.prefix)
        old_egg_info_file_md5 = compute_md5(fixer.egg_info_dir)
        fixer.repair()

        # Then
        assertCountEqual(self, os.listdir(fixer.egg_info_dir),
                         ["PKG-INFO", "egginst.json", "_info.json"])
        self.assertEqual(compute_md5(os.path.join(fixer.egg_info_dir, "PKG-INFO")),
                         old_egg_info_file_md5)

    @slow
    def test_repair(self):
        # Given
        broken_as_file_egg = DUMMY_EGG
        self._install_egg_file_egg_info_dir(broken_as_file_egg)

        broken_as_empty_dir = DUMMY_EGG_WITH_ENTRY_POINTS
        self._install_egg_empty_egg_info_dir(broken_as_empty_dir)

        non_broken_egg = DUMMY_EGG_WITH_APPINST
        self._install_egg(non_broken_egg)

        # When
        repair(self.prefix, False)

        # Then
        assertCountEqual(self, os.listdir(self._egg_info_path(broken_as_file_egg)),
                         ["PKG-INFO", "egginst.json", "_info.json"])
        assertCountEqual(self, os.listdir(self._egg_info_path(broken_as_empty_dir)),
                         ["entry_points.txt", "PKG-INFO", "egginst.json",
                          "_info.json"])
        assertCountEqual(self, os.listdir(self._egg_info_path(non_broken_egg)),
                         ["PKG-INFO"])

    @slow
    def test_repair_dry_run(self):
        # Given
        broken_as_file_egg = DUMMY_EGG
        self._install_egg_file_egg_info_dir(broken_as_file_egg)

        broken_as_empty_dir = DUMMY_EGG_WITH_ENTRY_POINTS
        self._install_egg_empty_egg_info_dir(broken_as_empty_dir)

        non_broken_egg = DUMMY_EGG_WITH_APPINST
        self._install_egg(non_broken_egg)

        # When
        repair(self.prefix, True)

        # Then
        self.assertTrue(os.path.isfile(self._egg_info_path(broken_as_file_egg)))
        assertCountEqual(self,
                         os.listdir(self._egg_info_path(broken_as_empty_dir)),
                         [])
        assertCountEqual(self, os.listdir(self._egg_info_path(non_broken_egg)),
                         ["PKG-INFO"])



import re

from egginst.vendor.six.moves import urllib

from enstaller.errors import InvalidConfiguration


_DEFAULT_PORT = 3128
_PORT_PROG_R = re.compile('^(.*):([0-9]+)$')
_PASSWD_PROG_R = re.compile('^([^:]*):(.*)$', re.S)


class ProxyInfo(object):
    @classmethod
    def from_string(cls, s):
        parts = urllib.parse.urlparse(s)
        if len(parts.scheme) > 0:
            scheme = parts.scheme
        else:
            scheme = "http"

        if len(parts.netloc) == 0:
            if len(parts.path) > 0:
                # this is to support url such as 'acme.com:3128'
                netloc = parts.path
            else:
                msg = "Invalid proxy string {0!r} (no host found)".format(s)
                raise InvalidConfiguration(msg)
        else:
            netloc = parts.netloc

        userpass, hostport = urllib.parse.splituser(netloc)
        if userpass is None:
            user, password = "", ""
        else:
            user, password = _splitpasswd(userpass)
        host, port = _splitport(hostport)
        if port is None:
            port = _DEFAULT_PORT
        else:
            port = int(port)

        if len(host) == 0:
            msg = "Invalid proxy string {0!r} (no host found)"
            raise InvalidConfiguration(msg.format(s))

        return cls(host, scheme, port, user, password)

    def __init__(self, host, scheme="http", port=_DEFAULT_PORT, user=None,
                 password=None):
        self._host = host
        self._scheme = scheme
        self._port = port
        self._user = user or ""
        self._password = password or ""

        if self._password and not self._user:
            msg = "One cannot create a proxy setting with a password but " \
                  "without a user "
            raise InvalidConfiguration(msg)

    def __str__(self):
        netloc = "{0}:{1}".format(self.host, self.port)

        if self.user:
            netloc = "{0}:{1}@{2}".format(self.user, self.password, netloc)

        return urllib.parse.urlunparse((self.scheme, netloc, "", "", "", ""))

    @property
    def host(self):
        return self._host

    @property
    def password(self):
        return self._password

    @property
    def port(self):
        return self._port

    @property
    def scheme(self):
        return self._scheme

    @property
    def user(self):
        return self._user


# Looks like those are semi-private function in the stdlib, so we bundle it
# here for 2/3 compat
def _splitport(host):
    match = _PORT_PROG_R.match(host)
    if match:
        return match.group(1, 2)
    else:
        return host, None


def _splitpasswd(user):
    match = _PASSWD_PROG_R.match(user)
    if match:
        return match.group(1, 2)
    else:
        return user, None



from egginst.vendor.six.moves import unittest

from ..pep386_workaround import PEP386WorkaroundVersion


class TestPEP386Workaround(unittest.TestCase):
    def test_correct_equal(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("1.3.0")
        right = PEP386WorkaroundVersion.from_string("1.3.0")

        # When/Then
        self.assertEqual(left, right)
        self.assertTrue(left == right)
        self.assertFalse(left != right)
        self.assertTrue(left <= right)
        self.assertTrue(left >= right)
        self.assertFalse(left < right)
        self.assertFalse(left > right)

        self.assertIs(left.is_worked_around, False)
        self.assertIs(right.is_worked_around, False)

    def test_hashing(self):
        # Given version string without workaround
        s = "1.3.0"

        # When
        v1 = PEP386WorkaroundVersion.from_string(s)
        v2 = PEP386WorkaroundVersion.from_string(s)

        # Then
        self.assertEqual(v1, v2)
        self.assertEqual(hash(v1), hash(v2))

        # Given version string with workaround
        s = "0.9.8j"

        # When
        v1 = PEP386WorkaroundVersion.from_string(s)
        v2 = PEP386WorkaroundVersion.from_string(s)

        # Then
        self.assertEqual(v1, v2)
        self.assertEqual(hash(v1), hash(v2))

    def test_correct_greater(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("1.2.0")
        right = PEP386WorkaroundVersion.from_string("1.3.0")

        # When/Then
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertTrue(left <= right)
        self.assertFalse(left >= right)
        self.assertTrue(left < right)
        self.assertFalse(left > right)

        # Given
        left = PEP386WorkaroundVersion.from_string("1.3.0rc2")
        right = PEP386WorkaroundVersion.from_string("1.3.0")

        # When/Then
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertTrue(left <= right)
        self.assertFalse(left >= right)
        self.assertTrue(left < right)
        self.assertFalse(left > right)

    def test_correct_lesser(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("1.3.0")
        right = PEP386WorkaroundVersion.from_string("1.2.0")

        # When/Then
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertFalse(left <= right)
        self.assertTrue(left >= right)
        self.assertFalse(left < right)
        self.assertTrue(left > right)

        # Given
        left = PEP386WorkaroundVersion.from_string("1.3.0")
        right = PEP386WorkaroundVersion.from_string("1.3.0rc2")

        # When/Then
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertFalse(left <= right)
        self.assertTrue(left >= right)
        self.assertFalse(left < right)
        self.assertTrue(left > right)

    def test_correct_series(self):
        # Given
        version_strings = ("1.1.0", "1.2.0.dev1", "1.2.0a1", "1.2.0a2", "1.2.0b1", "1.2.0c1",
                           "1.2.0rc2", "1.2.0")
        versions = tuple(PEP386WorkaroundVersion.from_string(v) for v in
                         version_strings)

        # When/Then
        self.assertTrue(versions[0] < versions[1] < versions[2] < versions[3] <
                        versions[4] < versions[5] < versions[6] < versions[7])
        self.assertTrue(versions[7] > versions[6] > versions[5] > versions[4] >
                        versions[3] > versions[2] > versions[1] > versions[0])

    def test_incorrect_equal(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("2010k")
        right = PEP386WorkaroundVersion.from_string("2010k")

        # When/Then
        self.assertEqual(left, right)
        self.assertTrue(left == right)
        self.assertFalse(left != right)
        self.assertTrue(left <= right)
        self.assertTrue(left >= right)
        self.assertFalse(left < right)
        self.assertFalse(left > right)
        self.assertIs(left.is_worked_around, True)
        self.assertIs(right.is_worked_around, True)

    def test_incorrect_greater(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("2010i")
        right = PEP386WorkaroundVersion.from_string("2010k")

        # When/Then
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertTrue(left <= right)
        self.assertFalse(left >= right)
        self.assertTrue(left < right)
        self.assertFalse(left > right)

    def test_incorrect_lesser(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("2010k")
        right = PEP386WorkaroundVersion.from_string("2010i")

        # When/Then
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertFalse(left <= right)
        self.assertTrue(left >= right)
        self.assertFalse(left < right)
        self.assertTrue(left > right)

    def test_correct_incorrect(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("1.3.0")
        right = PEP386WorkaroundVersion.from_string("1.4.0k")

        # When/Then
        self.assertNotEqual(left, right)
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertFalse(left <= right)
        self.assertTrue(left >= right)
        self.assertFalse(left < right)
        self.assertTrue(left > right)

        # Given
        left = PEP386WorkaroundVersion.from_string("1.4.0k")
        right = PEP386WorkaroundVersion.from_string("1.3.0")

        # When/Then
        self.assertNotEqual(left, right)
        self.assertFalse(left == right)
        self.assertTrue(left != right)
        self.assertTrue(left <= right)
        self.assertFalse(left >= right)
        self.assertTrue(left < right)
        self.assertFalse(left > right)

    def test_roundtrip_workedaround(self):
        # Given
        version_string = "2011g"

        # When
        version = PEP386WorkaroundVersion.from_string(version_string)

        # Then
        self.assertTrue(version._is_worked_around)
        self.assertEqual(str(version), version_string)

        # Given
        version_string = "0.9.0rc2"

        # When
        version = PEP386WorkaroundVersion.from_string(version_string)

        # Then
        self.assertFalse(version._is_worked_around)
        self.assertEqual(str(version), version_string)


class TestTrailingZeros(unittest.TestCase):
    def test_leading_zeros(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("0.9.8")
        right = PEP386WorkaroundVersion.from_string("1.0.1")

        # When/Then
        self.assertTrue(left < right)
        self.assertTrue(left <= right)
        self.assertFalse(left > right)
        self.assertFalse(left >= right)

    def test_round_trip(self):
        # Given
        version_string = "1.3.0"

        # When
        version = PEP386WorkaroundVersion.from_string(version_string)

        # Then
        self.assertEqual(str(version), version_string)

    def test_simple(self):
        # Given
        left = PEP386WorkaroundVersion.from_string("1.3.0")
        right = PEP386WorkaroundVersion.from_string("1.3")

        # When/Then
        self.assertEqual(left, right)
        self.assertTrue(left == right)
        self.assertFalse(left < right)
        self.assertTrue(left <= right)
        self.assertFalse(left > right)
        self.assertTrue(left >= right)

        # Given
        left = PEP386WorkaroundVersion.from_string("1.0.0")
        right = PEP386WorkaroundVersion.from_string("1")

        # When/Then
        self.assertEqual(left, right)
        self.assertTrue(left == right)
        self.assertFalse(left < right)
        self.assertTrue(left <= right)
        self.assertFalse(left > right)
        self.assertTrue(left >= right)

    def test_all_zeros(self):
        # Given
        version_string = "0.0.0"

        # When
        v = PEP386WorkaroundVersion.from_string(version_string)

        # Then
        self.assertEqual(str(v), version_string)


class TestOldCases(unittest.TestCase):
    def test_single_number(self):
        # Given
        version_string = "214"
        r_version = PEP386WorkaroundVersion.from_string("214.0.0")

        # When
        version = PEP386WorkaroundVersion.from_string(version_string)

        # Then
        self.assertEqual(version, r_version)
        self.assertEqual(str(version), version_string)

    def test_dev_tag_without_number(self):
        # Given
        r_version = PEP386WorkaroundVersion.from_string("1.0.0.dev1")
        version_string = "1.0.0.dev"

        # When
        version = PEP386WorkaroundVersion.from_string(version_string)

        # Then
        self.assertEqual(version, r_version)



from egginst.vendor.six.moves import unittest

from enstaller.repository import Repository
from enstaller.tests.common import dummy_repository_package_factory

from ..requirement import Requirement
from ..resolve import Resolve


class TestResolve(unittest.TestCase):
    def _repository_factory(self, packages):
        repository = Repository()
        for p in packages:
            repository.add_package(p)
        return repository

    def test__latest_egg_simple(self):
        # Given
        packages = [
            dummy_repository_package_factory("swig", "1.3.40", 1),
            dummy_repository_package_factory("swig", "1.3.40", 2),
            dummy_repository_package_factory("swig", "2.0.1", 1),
        ]
        repository = self._repository_factory(packages)

        # When
        resolver = Resolve(repository)
        latest = resolver._latest_egg(Requirement("swig"))

        # Then
        self.assertEqual(latest, "swig-2.0.1-1.egg")

        # When
        resolver = Resolve(repository)
        latest = resolver._latest_egg(Requirement("swigg"))

        # Then
        self.assertIsNone(latest)

    def test__latest_egg_multiple_python_versions(self):
        # Given
        packages = [
            dummy_repository_package_factory("swig", "1.3.40", 1),
            dummy_repository_package_factory("swig", "1.3.40", 2),
            dummy_repository_package_factory("swig", "2.0.1", 1)
        ]
        repository = self._repository_factory(packages)

        # When
        resolver = Resolve(repository)
        latest = resolver._latest_egg(Requirement("swig"))

        # Then
        self.assertEqual(latest, "swig-2.0.1-1.egg")



from os.path import dirname, join, isfile
from os import unlink

from egginst.vendor.six.moves import unittest

from enstaller.history import History


PATH = join(dirname(__file__), 'history')


def package_changes(first_set, second_set):
    removed = ['-' + name for name in first_set - second_set]
    added = ['+' + name for name in second_set - first_set]
    return set(removed + added)


class TestHistory(unittest.TestCase):

    def setUp(self):
        self.history = History('<dummy prefix>')
        self.history._log_path = PATH
        self.package_sets = [set(['appinst-2.1.0-1.egg',
                                  'basemap-1.0.1-1.egg',
                                  'biopython-1.57-2.egg']),

                             set(['basemap-1.0.2-1.egg',
                                  'biopython-1.57-2.egg',
                                  'numpy-1.6.1-2.egg']),

                             set(['basemap-1.0.2-1.egg',
                                  'biopython-1.57-2.egg',
                                  'numpy-1.7.0-1.egg'])]
        # Write the history file
        self.history._write_egg_names(self.package_sets[0])
        for i, state in enumerate(self.package_sets[:-1]):
            self.history._write_changes(state, self.package_sets[i+1])

    def tearDown(self):
        if isfile(self.history._log_path):
            unlink(self.history._log_path)

    def test_get_state(self):
        self.assertEqual(self.history.get_state(0), self.package_sets[0])
        self.assertEqual(self.history.get_state(), self.package_sets[-1])

    def test_parse(self):
        package_edits = [packages for (date, packages) in self.history.parse()]
        self.assertEqual(self.package_sets[0], package_edits[0])
        self.assertEqual(package_changes(self.package_sets[0],
                                         self.package_sets[1]),
                         package_edits[1])
        self.history.print_log()

        # When history file is not present
        unlink(self.history._log_path)
        self.assertEqual(self.history.parse(), [])
        self.history.print_log()

    def test_update(self):
        self.history.update()
        self.assertEqual(self.history.get_state(), set())
        self.assertEqual(self.history.parse()[-1][1],
                         package_changes(self.package_sets[-1], set()))

if __name__ == '__main__':
    unittest.main()



# THIS FILE IS GENERATED FROM ENSTALLER SETUP.PY
version = '4.8.7'
full_version = '4.8.7'
git_revision = '4ba6fba3034936c6cfcacd878a7cff3c81dda174'
is_released = True

if not is_released:
    version = full_version



"""
Simple py2/py3 shim
"""
import logging


# For compatibility with 2.6
class NullHandler(logging.Handler):  # pragma: no cover
    def emit(self, record):
        pass

import egginst.vendor.six

egginst.vendor.six.add_move(egginst.vendor.six.MovedModule("unittest",
                                                           "unittest2",
                                                           "unittest"))

PY2 = egginst.vendor.six.PY2

if PY2:
    buffer = buffer
else:
    buffer = memoryview


def assertCountEqual(self, first, second, msg=None):
    if PY2:
        return self.assertItemsEqual(first, second, msg)
    else:
        return self.assertCountEqual(first, second, msg)


def input(prompt):
    # XXX: is defined as a function so that mock.patch can patch it without
    # trouble.
    if PY2:
        return raw_input(prompt)
    else:
        import builtins
        return builtins.input(prompt)


# Code taken from Jinja2
def with_metaclass(meta, *bases):
    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instanciation that replaces
    # itself with the actual metaclass.  Because of internal type checks
    # we also need to make sure that we downgrade the custom metaclass
    # for one level to something closer to type (that's why __call__ and
    # __init__ comes back from type etc.).
    #
    # This has the advantage over six.with_metaclass in that it does not
    # introduce dummy classes into the final MRO.
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})



import re
import sys
import time

from os.path import isfile, join

import egginst


TIME_FMT = '%Y-%m-%d %H:%M:%S %z %Z'


def now():
    """
    return the current local time as an ISO formated
    string with time zone data, e.g. '2012-05-21 16:45:59 -0500 CDT'
    """
    return time.strftime(TIME_FMT)


def is_diff(content):
    return any(s.startswith(('-', '+')) for s in content)


def pretty_diff(diff):
    added = {}
    removed = {}
    for s in diff:
        fn = s[1:]
        name, version = egginst.name_version_fn(fn)
        if s.startswith('-'):
            removed[name.lower()] = version
        elif s.startswith('+'):
            added[name.lower()] = version
    changed = set(added) & set(removed)
    for name in sorted(changed):
        yield ' %s  (%s -> %s)' % (name, removed[name], added[name])
    for name in sorted(set(removed) - changed):
        yield '-%s-%s' % (name, removed[name])
    for name in sorted(set(added) - changed):
        yield '+%s-%s' % (name, added[name])


def pretty_content(content):
    if is_diff(content):
        return pretty_diff(content)
    else:
        return iter(sorted(content, key=lambda s: s.lower()))


class History(object):

    def __init__(self, prefix):
        self.prefix = prefix
        if prefix is None:
            return
        self._log_path = join(prefix, 'enpkg.hist')

    def __enter__(self):
        if self.prefix is None:
            return
        self.update()

    def __exit__(self, exc_type, exc_value, traceback):
        if self.prefix is None:
            return
        self.update()

    def _init_log_file(self, force=False):
        """
        initialize the log file
        """
        if not force and isfile(self._log_path):
            return
        self._write_egg_names(egginst.get_installed(self.prefix))

    def update(self):
        """
        update the history file (creating a new one if necessary)
        """
        self._init_log_file()
        last = self.get_state()
        curr = set(egginst.get_installed(self.prefix))
        if last == curr:
            return
        self._write_changes(last, curr)

    def parse(self):
        """
        parse the history file and return a list of
        tuples(datetime strings, set of eggs/diffs)
        """
        res = []
        if not isfile(self._log_path):
            return res
        sep_pat = re.compile(r'==>\s*(.+?)\s*<==')
        for line in open(self._log_path):
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            m = sep_pat.match(line)
            if m:
                dt = m.group(1)
                res.append((dt, set()))
            else:
                res[-1][1].add(line)
        return res

    def construct_states(self):
        """
        return a list of tuples(datetime strings, set of eggs)
        """
        res = []
        for dt, cont in self.parse():
            if not is_diff(cont):
                cur = cont
            else:
                for s in cont:
                    if s.startswith('-'):
                        cur.discard(s[1:])
                    elif s.startswith('+'):
                        cur.add(s[1:])
                    else:
                        raise Exception('Did not expect: %s' % s)
            res.append((dt, cur.copy()))
        return res

    def get_state(self, rev=-1):
        """
        return the state, i.e. the set of eggs, for a given revision,
        defaults to latest (which is the same as the current state when
        the log file is up-to-date)
        """
        times, pkgs = zip(*self.construct_states())
        return pkgs[rev]

    def print_log(self):
        for i, (date, content) in enumerate(self.parse()):
            print('%s  (rev %d)' % (date, i))
            for line in pretty_content(content):
                print('    %s' % line)
            print()

    def _write_egg_names(self, names):
        """ update the log file with the given list of egg names.

        Should be used only when init-ing the log_file
        """
        fo = open(self._log_path, 'w')
        fo.write("==> %s <==\n" % now())
        for eggname in names:
            fo.write('%s\n' % eggname)
        fo.close()

    def _write_changes(self, last_state, current_state):
        """ write the changes between last_state and current_state to log_file.

        """
        fo = open(self._log_path, 'a')
        fo.write("==> %s <==\n" % now())
        for fn in last_state - current_state:
            fo.write('-%s\n' % fn)
        for fn in current_state - last_state:
            fo.write('+%s\n' % fn)
        fo.close()


if __name__ == '__main__':
    h = History(sys.prefix)
    with h:
        h.update()
        h.print_log()



from binascii import hexlify, unhexlify
from hashlib import md5, sha1

from ..exceptions import SSLError


try:  # Test for SSL features
    SSLContext = None
    HAS_SNI = False

    import ssl
    from ssl import wrap_socket, CERT_NONE, PROTOCOL_SSLv23
    from ssl import SSLContext  # Modern SSL?
    from ssl import HAS_SNI  # Has SNI?
except ImportError:
    pass


def assert_fingerprint(cert, fingerprint):
    """
    Checks if given fingerprint matches the supplied certificate.

    :param cert:
        Certificate as bytes object.
    :param fingerprint:
        Fingerprint as string of hexdigits, can be interspersed by colons.
    """

    # Maps the length of a digest to a possible hash function producing
    # this digest.
    hashfunc_map = {
        16: md5,
        20: sha1
    }

    fingerprint = fingerprint.replace(':', '').lower()

    digest_length, rest = divmod(len(fingerprint), 2)

    if rest or digest_length not in hashfunc_map:
        raise SSLError('Fingerprint is of invalid length.')

    # We need encode() here for py32; works on py2 and p33.
    fingerprint_bytes = unhexlify(fingerprint.encode())

    hashfunc = hashfunc_map[digest_length]

    cert_digest = hashfunc(cert).digest()

    if not cert_digest == fingerprint_bytes:
        raise SSLError('Fingerprints did not match. Expected "{0}", got "{1}".'
                       .format(hexlify(fingerprint_bytes),
                               hexlify(cert_digest)))


def resolve_cert_reqs(candidate):
    """
    Resolves the argument to a numeric constant, which can be passed to
    the wrap_socket function/method from the ssl module.
    Defaults to :data:`ssl.CERT_NONE`.
    If given a string it is assumed to be the name of the constant in the
    :mod:`ssl` module or its abbrevation.
    (So you can specify `REQUIRED` instead of `CERT_REQUIRED`.
    If it's neither `None` nor a string we assume it is already the numeric
    constant which can directly be passed to wrap_socket.
    """
    if candidate is None:
        return CERT_NONE

    if isinstance(candidate, str):
        res = getattr(ssl, candidate, None)
        if res is None:
            res = getattr(ssl, 'CERT_' + candidate)
        return res

    return candidate


def resolve_ssl_version(candidate):
    """
    like resolve_cert_reqs
    """
    if candidate is None:
        return PROTOCOL_SSLv23

    if isinstance(candidate, str):
        res = getattr(ssl, candidate, None)
        if res is None:
            res = getattr(ssl, 'PROTOCOL_' + candidate)
        return res

    return candidate


if SSLContext is not None:  # Python 3.2+
    def ssl_wrap_socket(sock, keyfile=None, certfile=None, cert_reqs=None,
                        ca_certs=None, server_hostname=None,
                        ssl_version=None):
        """
        All arguments except `server_hostname` have the same meaning as for
        :func:`ssl.wrap_socket`

        :param server_hostname:
            Hostname of the expected certificate
        """
        context = SSLContext(ssl_version)
        context.verify_mode = cert_reqs

        # Disable TLS compression to migitate CRIME attack (issue #309)
        OP_NO_COMPRESSION = 0x20000
        context.options |= OP_NO_COMPRESSION

        if ca_certs:
            try:
                context.load_verify_locations(ca_certs)
            # Py32 raises IOError
            # Py33 raises FileNotFoundError
            except Exception as e:  # Reraise as SSLError
                raise SSLError(e)
        if certfile:
            # FIXME: This block needs a test.
            context.load_cert_chain(certfile, keyfile)
        if HAS_SNI:  # Platform-specific: OpenSSL with enabled SNI
            return context.wrap_socket(sock, server_hostname=server_hostname)
        return context.wrap_socket(sock)

else:  # Python 3.1 and earlier
    def ssl_wrap_socket(sock, keyfile=None, certfile=None, cert_reqs=None,
                        ca_certs=None, server_hostname=None,
                        ssl_version=None):
        return wrap_socket(sock, keyfile=keyfile, certfile=certfile,
                           ca_certs=ca_certs, cert_reqs=cert_reqs,
                           ssl_version=ssl_version)



"""
This is the indexed_repo API
============================

FIXME: more text here

"""
from __future__ import absolute_import

from .chain import Chain
from .requirement import (Req, spec_as_req, filename_as_req, dist_as_req,
                          add_Reqs_to_spec)
from .metadata import spec_from_dist, parse_data
from .dist_naming import filename_dist, repo_dist



import re
from os.path import isdir

from enstaller.utils import abs_expanduser, comparable_version


DIST_PAT = re.compile(r'(file://.*[\\/]|https?://.+/)([^\\/]+)$')

def split_dist(dist):
    """
    splits a distribution, e.g. 'http://www.example.com/repo/foo.egg', into
    repo and filename ('http://www.example.com/repo/', 'foo.egg').

    A distribution string, usually named 'dist', is always repo + filename.
    That is, simply adding the two strings will give the dist.  The terms
    filename and distname are used interchangeably.  There are currently
    the three types of repos:

    file://
    =======

    This repository type refers to a directory on a local filesystem, and
    may or may not be indexed.  That is, when an index file is found it is
    used to add the repository to the index, otherwise (if no index file
    exists), the distributions are added to the index by inspecting the
    actual files.

    http:// or https://
    ===================

    A remote repository, which must contain a compressed index file.


    Naming examples:
    ================

    Here are some valid repo names:

    file:///usr/local/repo/
    file://E:\eggs\
    http://download.enthought.com/pypieggs/Windows/amd64/
    https://www.enthought.com/repo/epd/eggs/Windows/x86/
    https://username:password@www.enthought.com/repo/epd/eggs/Windows/x86/

    Note that, since we always have dist = repo + filename, the file:// repo
    name has to end with a forward slash (backslash on Windows), and the
    http:// always ends with a forward slash.
    """
    m = DIST_PAT.match(dist)
    assert m is not None, dist
    repo, filename = m.group(1), m.group(2)
    return repo, filename


def repo_dist(dist):
    return split_dist(dist)[0]


def filename_dist(dist):
    return split_dist(dist)[1]


egg_pat = re.compile(r'([\w.]+)-([\w.]+)-(\d+)\.egg$')

def is_valid_eggname(eggname):
    return bool(egg_pat.match(eggname))

def split_eggname(eggname):
    m = egg_pat.match(eggname)
    assert m, eggname
    return m.group(1), m.group(2), int(m.group(3))


def dirname_repo(repo):
    if repo.startswith('file://'):
        return repo[7:].rstrip(r'\/')
    return None


def cleanup_reponame(repo):
    """
    Make sure a given repo string, i.e. a string specifying a repository,
    is valid and return a cleaned up version of the string.
    """
    if repo.startswith(('http://', 'https://')):
        if not repo.endswith('/'):
            repo += '/'

    elif repo.startswith('file://'):
        dir_path = repo[7:]
        if dir_path.startswith('/'):
            # Unix filename
            if not repo.endswith('/'):
                repo += '/'
        else:
            # Windows filename
            if not repo.endswith('\\'):
                repo += '\\'

    elif isdir(abs_expanduser(repo)):
        return cleanup_reponame('file://' + abs_expanduser(repo))

    else:
        raise Exception("Invalid repo string or non-existing file: %r" % repo)

    return repo


def comparable_spec(spec):
    """
    Returns a tuple(version, build) for a distribution, version is a
    RationalVersion object.  The result may be used for as a sort key.
    """
    return comparable_version(spec['version']), spec['build']



from egginst.vendor.six.moves import urllib


def close_file_or_response(fp):
    if hasattr(fp, "close"):
        fp.close()
    else:
        # Compat shim for requests < 2
        fp._fp.close()


def path_to_uri(path):
    """Convert the given path to a file:// uri."""
    return urllib.parse.urljoin("file:", urllib.request.pathname2url(path))



import json
import os.path
import shutil
import tempfile

from egginst.vendor.six.moves import unittest
from egginst.tests.common import _EGGINST_COMMON_DATA

from enstaller.egg_meta import info_from_egg, update_index


class TestInfoFromEgg(unittest.TestCase):
    def test_simple(self):
        # Given
        r_info = {
            'arch': 'amd64',
            'build': 1,
            'name': 'nose',
            'osdist': None,
            'packages': [],
            'platform': 'darwin',
            'python': '2.7',
            'type': 'egg',
            'version': '1.3.0'
        }
        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-1.egg")

        # When
        info = info_from_egg(path)

        # Then
        self.assertEqual(info, r_info)


class TestUpdateIndex(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_simple(self):
        # Given
        for egg in ["nose-1.3.0-1.egg", "nose-1.3.0-2.egg"]:
            shutil.copy(os.path.join(_EGGINST_COMMON_DATA, egg), self.tempdir)
        index_path = os.path.join(self.tempdir, "index.json")

        # When
        update_index(self.tempdir)

        # Then
        with open(index_path) as fp:
            data = json.load(fp)
            print(data)



import logging
from collections import defaultdict

from enstaller.errors import MissingDependency, NoPackageFound
from enstaller.repository import egg_name_to_name_version

from .requirement import Requirement

logger = logging.getLogger(__name__)


class Resolve(object):
    """
    The main purpose of this class is to support the install_sequence method
    below.  In most cases, the user will only create an instance of this
    class (which is inexpensive), to call the install_sequence method, e.g.:

    eggs = Resolve(repository).install_sequence(req)
    """
    def __init__(self, repository):
        """
        Create a new Resolve instance

        Parameters
        ----------
        repository: repository
            The repository instance to use to query package metadata
        """
        self.repository = repository

    def _latest_egg(self, requirement):
        """
        return the egg with the largest version and build number
        """
        assert requirement.strictness >= 1
        d = dict((package.key, package) for package in
                 self.repository.find_packages(requirement.name))
        matches = [key for key, package in d.items()
                   if requirement.matches(package)]
        if len(matches) == 0:
            return None
        else:
            return max(matches, key=lambda k: d[k].version)

    def _dependencies_from_egg(self, egg):
        """
        return the set of requirement objects listed by the given egg
        """
        name, version = egg_name_to_name_version(egg)
        package = self.repository.find_package(name, version)
        return set(Requirement(s) for s in package.dependencies)

    def _name_from_egg(self, egg):
        """
        return the project name for a given egg (from it's meta data)
        """
        name, version = egg_name_to_name_version(egg)
        return self.repository.find_package(name, version).name

    def are_complete(self, eggs):
        """
        return True if the 'eggs' are complete, i.e. the for each egg all
        dependencies (by name only) are also included in 'eggs'
        """
        names = set(self._name_from_egg(d) for d in eggs)
        for egg in eggs:
            for r in self._dependencies_from_egg(egg):
                if r.name not in names:
                    return False
        return True

    def _determine_install_order(self, eggs):
        """
        given the 'eggs' (which are already complete, i.e. the for each
        egg all dependencies are also included in 'eggs'), return a list
        of the same eggs in the correct install order
        """
        eggs = list(eggs)
        assert self.are_complete(eggs)

        # make sure each project name is listed only once
        assert len(eggs) == len(set(self._name_from_egg(d) for d in eggs))

        # the eggs corresponding to the requirements must be sorted
        # because the output of this function is otherwise not deterministic
        eggs.sort(key=self._name_from_egg)

        # maps egg -> set of required (project) names
        rns = {}
        for egg in eggs:
            rns[egg] = set(r.name for r in self._dependencies_from_egg(egg))

        # as long as we have things missing, simply look for things which
        # can be added, i.e. all the requirements have been added already
        result = []
        names_inst = set()
        while len(result) < len(eggs):
            n = len(result)
            for egg in eggs:
                if egg in result:
                    continue
                # see if all required packages were added already
                if all(bool(name in names_inst) for name in rns[egg]):
                    result.append(egg)
                    names_inst.add(self._name_from_egg(egg))
                    assert len(names_inst) == len(result)

            if len(result) == n:
                # nothing was added
                raise Exception("Loop in dependency graph\n%r" % eggs)
        return result

    def _sequence_flat(self, root):
        eggs = [root]
        for r in self._dependencies_from_egg(root):
            d = self._latest_egg(r)
            if d is None:
                from enstaller.enpkg import EnpkgError
                err = EnpkgError('Error: could not resolve %s' % str(r))
                err.req = r
                raise err
            eggs.append(d)

        can_order = self.are_complete(eggs)
        logger.info("Can determine install order: %r", can_order)
        if can_order:
            eggs = self._determine_install_order(eggs)
        return eggs

    def _sequence_recur(self, root):
        reqs_shallow = {}
        for r in self._dependencies_from_egg(root):
            reqs_shallow[r.name] = r
        reqs_deep = defaultdict(set)

        def add_dependents(egg, visited=None):
            if visited is None:
                visited = set()
            visited.add(egg)
            for r in self._dependencies_from_egg(egg):
                reqs_deep[r.name].add(r)
                if (r.name in reqs_shallow and
                        r.strictness < reqs_shallow[r.name].strictness):
                    continue
                d = self._latest_egg(r)
                if d is None:
                    msg = "Could not resolve \"%s\" " \
                          "required by \"%s\"" % (str(r), egg)
                    raise MissingDependency(msg, egg, r)
                eggs.add(d)
                if d not in visited:
                    add_dependents(d, visited)

        eggs = set([root])
        add_dependents(root)

        names = set(self._name_from_egg(d) for d in eggs)
        if len(eggs) != len(names):
            for name in names:
                ds = [d for d in eggs if self._name_from_egg(d) == name]
                assert len(ds) != 0
                if len(ds) == 1:
                    continue
                logger.info('multiple: %s', name)
                for d in ds:
                    logger.info('    %s', d)
                r = max(reqs_deep[name], key=lambda r: r.strictness)
                assert r.name == name
                # remove the eggs with name
                eggs = [d for d in eggs if self._name_from_egg(d) != name]
                # add the one
                eggs.append(self._latest_egg(r))

        return self._determine_install_order(eggs)

    def install_sequence(self, req, mode='recur'):
        """
        Return the list of eggs which need to be installed (and None if
        the requirement can not be resolved).
        The returned list is given in dependency order.
        The 'mode' may be:

        'root':  only the egg for the requirement itself is
                 contained in the result (but not any dependencies)

        'flat':  dependencies are handled only one level deep

        'recur': dependencies are handled recursively (default)
        """
        logger.info("Determining install sequence for %r", req)
        root = self._latest_egg(req)
        if root is None:
            msg = "No egg found for requirement {0!r}.".format(str(req))
            raise NoPackageFound(msg, req)
        if mode == 'root':
            return [root]
        if mode == 'flat':
            return self._sequence_flat(root)
        if mode == 'recur':
            return self._sequence_recur(root)
        raise Exception('did not expect: mode = %r' % mode)



import json
import os.path
import platform
import shutil
import sys
import tempfile
import textwrap

import mock

from egginst.vendor.six.moves import unittest

from enstaller import __version__

from enstaller.config import _keyring_backend_name, Configuration
from enstaller.history import History
from enstaller.main import main_noexc
from enstaller.vendor import responses
from enstaller.utils import PY_VER

from enstaller.tests.common import mock_index, mock_print, R_JSON_AUTH_RESP

from .common import authenticated_config


class TestMisc(unittest.TestCase):
    @authenticated_config
    @responses.activate
    def test_print_config(self):
        self.maxDiff = None

        # Given
        config = Configuration()
        config.update(prefix=sys.prefix)

        template = textwrap.dedent("""\
Python version: {pyver}
enstaller version: {version}
sys.prefix: {sys_prefix}
platform: {platform}
architecture: {arch}
use_webservice: True
keyring backend: {keyring_backend}
settings:
    prefix = {prefix}
    repository_cache = {repository_cache}
    noapp = False
    proxy = None
You are logged in as 'dummy' (David Cournapeau).
Subscription level: Canopy / EPD Basic or above
""")
        r_output = template.format(pyver=PY_VER,
                                   sys_prefix=os.path.normpath(sys.prefix),
                                   version=__version__,
                                   platform=platform.platform(),
                                   arch=platform.architecture()[0],
                                   keyring_backend=_keyring_backend_name(),
                                   prefix=os.path.normpath(config.prefix),
                                   repository_cache=config.repository_cache)

        responses.add(responses.GET,
                      "https://api.enthought.com/accounts/user/info/",
                      body=json.dumps(R_JSON_AUTH_RESP))

        # When
        with self.assertRaises(SystemExit) as e:
            with mock_print() as m:
                main_noexc(["--config"])

        # Then
        self.assertEqual(e.exception.code, 0)
        self.assertMultiLineEqual(m.value, r_output)

    @authenticated_config
    def test_list_bare(self):
        # Given
        sys_prefix = os.path.normpath(sys.prefix)

        # When
        with mock.patch("enstaller.cli.commands.print_installed"):
            with self.assertRaises(SystemExit) as e:
                with mock_print() as m:
                    main_noexc(["--list"])

        # Then
        self.assertEqual(e.exception.code, 0)
        self.assertMultiLineEqual(m.value, "prefix: {0}\n\n".format(sys_prefix))

    @authenticated_config
    def test_log(self):
        with mock.patch("enstaller.cli.commands.History",
                        spec=History) as mocked_history:
            with self.assertRaises(SystemExit) as e:
                with mock_print() as m:
                    main_noexc(["--log"])
            self.assertEqual(e.exception.code, 0)
            self.assertTrue(mocked_history.return_value.print_log.called)
            self.assertMultiLineEqual(m.value, "")

    @authenticated_config
    def test_freeze(self):
        installed_requirements = ["dummy 1.0.0-1", "another_dummy 1.0.1-1"]
        with mock.patch("enstaller.cli.commands.get_freeze_list",
                        return_value=installed_requirements):
            with self.assertRaises(SystemExit) as e:
                with mock_print() as m:
                    main_noexc(["--freeze"])
            self.assertEqual(e.exception.code, 0)
            self.assertMultiLineEqual(m.value,
                                      "dummy 1.0.0-1\nanother_dummy 1.0.1-1\n")

    @mock_index({
        "fubar-1.0.0-1.egg": {
            "available": True,
            "build": 1,
            "md5": "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            "mtime": 0.0,
            "name": "fubar",
            "packages": [],
            "product": "nono",
            "python": PY_VER,
            "size": 0,
            "type": "egg",
            "version": "1.0.0"
            }}, "https://acme.com")
    def test_insecure_flag(self):
        # Given
        responses.add(responses.GET,
                      "https://acme.com/accounts/user/info/",
                      body=json.dumps(R_JSON_AUTH_RESP))

        config = Configuration()
        config.update(store_url="https://acme.com")
        config.update(auth=("nono", "le gros robot"))

        # When
        with self.assertRaises(SystemExit) as e:
            with mock.patch("enstaller.main._ensure_config_or_die",
                            return_value=config):
                with mock.patch("enstaller.main.convert_auth_if_required"):
                    main_noexc(["-s", "fubar"])

        # Then
        self.assertEqual(e.exception.code, 0)

        # When
        with self.assertRaises(SystemExit) as e:
            with mock.patch("enstaller.main._ensure_config_or_die",
                            return_value=config):
                with mock.patch("enstaller.main.convert_auth_if_required"):
                    main_noexc(["-ks", "fubar"])

        # Then
        self.assertEqual(e.exception.code, 0)


class TestPrefix(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    @authenticated_config
    @mock_index({
        "fubar-1.0.0-1.egg": {
            "available": True,
            "build": 1,
            "md5": "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            "mtime": 0.0,
            "name": "fubar",
            "packages": [],
            "product": "nono",
            "python": PY_VER,
            "size": 0,
            "type": "egg",
            "version": "1.0.0"
            }}, "https://api.enthought.com")
    def test_simple(self):
        self.maxDiff = None

        # Given
        responses.add(responses.GET,
                      "https://api.enthought.com/accounts/user/info/",
                      body=json.dumps(R_JSON_AUTH_RESP))

        template = textwrap.dedent("""\
Python version: {pyver}
enstaller version: {version}
sys.prefix: {sys_prefix}
platform: {platform}
architecture: {arch}
use_webservice: True
keyring backend: {keyring_backend}
settings:
    prefix = {prefix}
    repository_cache = {repository_cache}
    noapp = False
    proxy = None
You are logged in as 'dummy' (David Cournapeau).
Subscription level: Canopy / EPD Basic or above
""")
        r_output = template.format(pyver=PY_VER,
                                   sys_prefix=os.path.normpath(sys.prefix),
                                   version=__version__,
                                   platform=platform.platform(),
                                   arch=platform.architecture()[0],
                                   keyring_backend=_keyring_backend_name(),
                                   prefix=os.path.normpath(self.prefix),
                                   repository_cache=os.path.join(self.prefix,
                                                                 "LOCAL-REPO"))

        # When
        with self.assertRaises(SystemExit):
            with mock_print() as m:
                main_noexc(["--config", "--prefix={0}".format(self.prefix)])

        # Then
        self.assertEqual(m.value, r_output)



# Copyright by Enthought, Inc.
# Author: Ilan Schnell <ischnell@enthought.com>
from __future__ import absolute_import, print_function

import logging
import re
import os
import sys
import platform
import tempfile
import warnings

from os.path import isfile, join

from egginst.utils import parse_assignments
from egginst.vendor.six import string_types
from egginst.vendor.six.moves import urllib

from enstaller.vendor import keyring
from enstaller.vendor.keyring.backends.file import PlaintextKeyring

from enstaller import __version__
from enstaller.auth import (_INDEX_NAME, DUMMY_USER,
                            subscription_message, APITokenAuth, UserInfo,
                            UserPasswordAuth)
from enstaller.config_templates import RC_DEFAULT_TEMPLATE, RC_TEMPLATE
from enstaller.errors import (EnstallerException, InvalidConfiguration,
                              InvalidFormat)
from enstaller.proxy_info import ProxyInfo
from enstaller.utils import real_prefix, under_venv
from enstaller.vendor import requests
from enstaller.cli.utils import humanize_ssl_error_and_die
from enstaller import plat
from .utils import PY_VER, abs_expanduser, fill_url
from ._yaml_config import load_configuration_from_yaml

logger = logging.getLogger(__name__)

KEYRING_SERVICE_NAME = 'Enthought.com'

ENSTALLER4RC_FILENAME = ".enstaller4rc"
SYS_PREFIX_ENSTALLER4RC = os.path.join(sys.prefix, ENSTALLER4RC_FILENAME)
HOME_ENSTALLER4RC = os.path.join(abs_expanduser("~"), ENSTALLER4RC_FILENAME)

STORE_KIND_LEGACY = "legacy"
STORE_KIND_BROOD = "brood"
_BROOD_PREFIX = "brood+"


def _setup_keyring():
    backend = PlaintextKeyring()

    try:
        if sys.platform == "win32":
            from enstaller.vendor.keyring.backends.Windows import \
                WinVaultKeyring
            tentative_backend = WinVaultKeyring()
        elif sys.platform == "darwin":
            from enstaller.vendor.keyring.backends.OS_X import Keyring
            tentative_backend = Keyring()
        else:
            tentative_backend = backend
        if tentative_backend.priority >= 0:
            backend = tentative_backend
    except ImportError:
        pass

    keyring.set_keyring(backend)

_setup_keyring()


def _keyring_backend_name():
    return str(type(keyring.get_keyring()))


def legacy_configuration_read_search_order():
    """
    Return a list of directories where to look for the configuration file.

    Legacy order, don't use outside canopy.
    """
    paths = [
        abs_expanduser("~"),
        real_prefix(),
    ]

    return [os.path.normpath(p) for p in paths]


def configuration_read_search_order():
    """
    Return a list of directories where to look for the configuration file.
    """
    paths = [sys.prefix]
    paths.append(abs_expanduser("~"))
    if under_venv():
        paths.append(real_prefix())

    return [os.path.normpath(p) for p in paths]


def add_url(filename, config, url):
    url = fill_url(url)
    if url in config.indexed_repositories:
        print("Already configured:", url)
        return
    prepend_url(filename, url)


def _get_writable_local_dir(local_dir):
    if not os.access(local_dir, os.F_OK):
        try:
            os.makedirs(local_dir)
            return local_dir
        except (OSError, IOError):
            pass
    elif os.access(local_dir, os.W_OK):
        return local_dir

    logger.warn('Warning: the following directory is not writeable '
                'with current permissions:\n'
                '    {0!r}\n'
                'Using a temporary cache for index and eggs.\n'.
                format(local_dir))
    return tempfile.mkdtemp()


def write_default_config(filename):
    """
    Write a default configuration file at the given location.

    Will raise an exception if a file already exists.

    Parameters
    ----------
    filename : str
        The location to write to.
    """
    if os.path.isfile(filename):
        msg = "File '{0}' already exists, not overwriting."
        raise EnstallerException(msg.format(filename))
    else:
        config = Configuration()

        if config.proxy:
            proxy_line = 'proxy = %r' % str(config.proxy)
        else:
            proxy_line = ('#proxy = <proxy string>  '
                          '# e.g. "http://<user>:<passwd>@123.0.1.2:8080"')

        variables = {"py_ver": PY_VER, "sys_prefix": sys.prefix, "version":
                     __version__, "proxy_line": proxy_line,
                     "auth_section": ""}
        with open(filename, "w") as fo:
            fo.write(RC_DEFAULT_TEMPLATE % variables)


def _is_using_epd_username(filename_or_fp):
    """
    Returns True if the given configuration file uses EPD_username.
    """
    data = parse_assignments(filename_or_fp)
    return "EPD_username" in data and "EPD_auth" not in data


def convert_auth_if_required(filename):
    """
    This function will convert configuration using EPD_username + keyring to
    using EPD_auth.

    Returns True if the file has been modified, False otherwise.
    """
    did_convert = False
    if _is_using_epd_username(filename):
        msg = "Cannot convert password: no password found in keyring"
        config = Configuration.from_file(filename)
        if config.auth is None:
            raise EnstallerException(msg)
        username = config.auth.username
        password = _get_keyring_password(username)
        if password is None:
            raise EnstallerException(msg)
        else:
            config.update(auth=(username, password))
            config._change_auth(filename)
            did_convert = True

    return did_convert


def _get_keyring_password(username):
    return keyring.get_password(KEYRING_SERVICE_NAME, username)


def _set_keyring_password(username, password):
    return keyring.set_password(KEYRING_SERVICE_NAME, username, password)


def _create_error_message(fp, exc):
    pos = fp.tell()
    try:
        fp.seek(0)
        lines = fp.readlines()
        line = lines[exc.lineno-1].rstrip()
        if isinstance(exc, SyntaxError):
            msg = "Could not parse configuration file " \
                  "(invalid python syntax at line {0!r}: expression {1!r})".\
                  format(exc.lineno, line)
        else:
            msg = "Could not parse configuration file " \
                  "(error at line {0!r}: expression {1!r} not " \
                  "supported)".format(exc.lineno, line)
        return msg
    finally:
        fp.seek(pos)


class Configuration(object):
    @classmethod
    def _from_legacy_locations(cls):
        warnings.warn("Using legacy location is deprecated, use "
                      "Configuration.from_filename with an explicit "
                      "filename instead", DeprecationWarning)
        config_path = None
        for p in legacy_configuration_read_search_order():
            candidate = os.path.join(p, ENSTALLER4RC_FILENAME)
            if isfile(candidate):
                config_path = candidate
                break

        if config_path is None:
            raise InvalidConfiguration("No default configuration found.")
        else:
            return cls.from_file(config_path)

    @classmethod
    def from_yaml_filename(cls, filename):
        return load_configuration_from_yaml(cls, filename)

    @classmethod
    def from_file(cls, filename):
        """
        Create a new Configuration instance from the given file.

        Parameters
        ----------
        filename: str or file-like object
            If a string, is understood as a filename to open. Understood as a
            file-like object otherwise.
        """
        def _create(fp):
            ret = cls()

            def api_token_to_auth(api_token):
                ret.update(auth=APITokenAuth(api_token))

            def epd_auth_to_auth(epd_auth):
                ret.update(auth=UserPasswordAuth.from_encoded_auth(epd_auth))

            def epd_username_to_auth(username):
                if keyring is not None:
                    password = _get_keyring_password(username)
                    if password is not None:
                        ret.update(auth=(username, password))

            try:
                parsed = parse_assignments(fp)
            except (InvalidFormat, SyntaxError) as e:
                msg = _create_error_message(fp, e)
                raise InvalidConfiguration(msg)

            # We need a custom translator to manage attributes in the
            # configuration file without polluting the config class
            translator = ret._name_to_setter.copy()
            translator.update({
                "EPD_auth": epd_auth_to_auth,
                "EPD_username": epd_username_to_auth,
                "IndexedRepos": translator["indexed_repositories"],
                "api_token": api_token_to_auth,
            })

            if "EPD_auth" in parsed and "api_token" in parsed:
                msg = "Both 'EPD_auth' and 'api_token' set in configuration." \
                      "\nYou should remove one of those for consistent " \
                      "behaviour."
                warnings.warn(msg)

            for name, value in parsed.items():
                if name in translator:
                    translator[name](value)
                else:
                    warnings.warn("Unsupported configuration setting {0}, "
                                  "ignored".format(name))
            return ret

        if isinstance(filename, string_types):
            with open(filename, "r") as fp:
                ret = _create(fp)
                ret._filename = filename
                return ret
        else:
            return _create(filename)

    def __init__(self, **kw):
        """ Create a new configuration instance.

        Any argument passed to the constructor need to be a keyword argument,
        and is understood as an argument to the update method, e.g.::

            config = Configuration(store_url="http://acme.com")

            # equivalent to
            config = Configuration()
            config.update(store_url="http://acme.com")
        """
        self._auth = None
        self._autoupdate = True
        self._noapp = False
        self._proxy = None
        self._use_pypi = True
        self._use_webservice = True

        self._prefix = os.path.normpath(sys.prefix)
        self._indexed_repositories = []
        self._store_url = "https://api.enthought.com"
        self._store_kind = STORE_KIND_LEGACY

        self._repository_cache = None

        self._filename = None
        self._platform = plat.custom_plat

        self._max_retries = 0
        self._verify_ssl = True

        self._name_to_setter = {}
        simple_attributes = [
            ("autoupdate", "_autoupdate"),
            ("noapp", "_noapp"),
            ("verify_ssl", "_verify_ssl"),
            ("use_pypi", "_use_pypi"),
            ("use_webservice", "_use_webservice"),
        ]
        for name, private_attribute in simple_attributes:
            self._name_to_setter[name] = \
                self._simple_attribute_set_factory(private_attribute)

        self._name_to_setter.update({
            "auth": self._set_auth,
            "indexed_repositories": self._set_indexed_repositories,
            "max_retries": self._set_max_retries,
            "prefix": self._set_prefix,
            "proxy": self._set_proxy,
            "repository_cache": self._set_repository_cache,
            "store_url": self._set_store_url,
        })

        self.update(**kw)

    # ----------
    # Properties
    # ----------
    @property
    def api_url(self):
        """
        Url to hit to get user information on api.e.com.
        """
        return fill_url("{0}/accounts/user/info/".format(self.store_url))

    @property
    def auth(self):
        """
        The auth object that may be passed to Session.authenticate

        :return: the auth instance, or None is configured.
        :rtype: IAuth or None
        """
        return self._auth

    @property
    def autoupdate(self):
        """
        Whether enpkg should attempt updating itself.
        """
        return self._autoupdate

    @property
    def filename(self):
        """
        The filename this configuration was created from. May be None if the
        configuration was not created from a file.
        """
        return self._filename

    @property
    def indexed_repositories(self):
        """
        List of (old-style) repositories. Only actually used when
        use_webservice is False.
        """
        return self._indexed_repositories

    @property
    def indices(self):
        """
        Returns a list of pair (index_url, store_location) for this given
        configuration.

        Takes into account webservice/no webservice and pypi True/False
        """
        if self.use_webservice:
            index_url = store_url = self.webservice_entry_point + _INDEX_NAME
            if self.use_pypi:
                index_url += "?pypi=true"
            else:
                index_url += "?pypi=false"
            return tuple([(index_url, store_url)])
        else:
            return tuple((url + _INDEX_NAME, url + _INDEX_NAME)
                         for url in self.indexed_repositories)

    @property
    def max_retries(self):
        """
        Max attempts to retry an http connection or re-fetching data whose
        checksum failed.
        """
        return self._max_retries

    @property
    def noapp(self):
        """
        Ignore appinst entries.
        """
        return self._noapp

    @property
    def prefix(self):
        """
        Prefix in which enpkg operates.
        """
        return self._prefix

    @property
    def proxy(self):
        """
        A ProxyInfo instance or None if no proxy is configured.
        """
        return self._proxy

    @property
    def proxy_dict(self):
        """
        A dictionary <scheme>:<proxy_string> that can be used as the proxies
        argument for requests.
        """
        if self._proxy:
            return {self._proxy.scheme: str(self._proxy)}
        else:
            return {}

    @property
    def repository_cache(self):
        """
        Absolute path where eggs will be cached.
        """
        if self._repository_cache is None:
            return join(self.prefix, "LOCAL-REPO")
        else:
            return self._repository_cache

    @property
    def store_kind(self):
        """
        Store kind (brood, legacy canopy, old-repo style).
        """
        return self._store_kind

    @property
    def store_url(self):
        """
        The store url to hit for indices and eggs.
        """
        return self._store_url

    @property
    def verify_ssl(self):
        """
        Whether to verify SSL CA or not.
        """
        return self._verify_ssl

    @property
    def use_pypi(self):
        """
        Whether to load pypi repositories (in `webservice` mode).
        """
        return self._use_pypi

    @property
    def use_webservice(self):
        """
        Whether to use canopy legacy or not.
        """
        return self._use_webservice

    @property
    def webservice_entry_point(self):
        """
        Whether to fetch indices and data (in `webservice` mode).
        """
        return fill_url("{0}/eggs/{1}/".
                        format(self.store_url, self._platform))

    # --------------
    # Public methods
    # --------------
    def reset_auth(self):
        self._auth = None

    def set_auth_from_encoded(self, value):
        try:
            auth = UserPasswordAuth.from_encoded_auth(value)
        except Exception:
            raise InvalidConfiguration("Invalid EPD_auth value")
        else:
            self.update(auth=auth)

    def set_repositories_from_names(self, names):
        """ Set repositories from their names alone, e.g. 'enthought/free'.

        Parameters
        ----------
        names : list
            List of repository names
        """
        repository_urls = []
        for name in names:
            p = urllib.parse.urlparse(name)
            if p.scheme == "":
                url = self.store_url + "/repo/{0}/{{PLATFORM}}".format(name)
            elif p.scheme == "file":
                url = name
            else:
                msg = "Unsupported syntax: {0!r}".format(name)
                raise InvalidConfiguration(msg)
            repository_urls.append(url)
        self.update(indexed_repositories=repository_urls)

    def update(self, **kw):
        """ Set configuration attributes given as keyword arguments."""
        for name, value in kw.items():
            setter = self._name_to_setter.get(name, None)
            if name is None:
                raise ValueError("Invalid setting name: {0!r}".format(name))
            else:
                setter(value)

    def write(self, filename):
        """ Write this configuration to the given filename.

        Parameters
        ----------
        filename : str
            The path of the written file.
        """
        if self.auth is None:
            auth_section = ""
        else:
            auth_section = self.auth.config_string

        if self.proxy:
            proxy_line = 'proxy = %r' % str(self.proxy)
        else:
            proxy_line = ('#proxy = <proxy string>  '
                          '# e.g. "http://<user>:<passwd>@123.0.1.2:8080"')

        # XXX: because setting indexed repositories use string replacement
        # (sigh...), we need to write the IndexedRepos section in a very specific
        # way:
        #
        #     IndexedRepos = [
        #        url1,
        #        url2
        #    ]
        indexed_repos_string = "\n".join("    %r," % (repo,)
                                         for repo in self.indexed_repositories)

        if self.store_kind == STORE_KIND_BROOD:
            store_url = _BROOD_PREFIX + self.store_url
        else:
            store_url = self.store_url

        variables = {"py_ver": PY_VER, "sys_prefix": sys.prefix, "version":
                     __version__, "proxy_line": proxy_line, "auth_section":
                     auth_section, "store_url": store_url,
                     "indexed_repositories": indexed_repos_string,
                     "use_pypi": self.use_pypi,
                     "use_webservice": self.use_webservice,
                     }
        with open(filename, "w") as fo:
            fo.write(RC_TEMPLATE % variables)

    # ---------------
    # Private methods
    # ---------------
    def _change_auth(self, filename):
        if self.auth is None:
            pat = re.compile(r'^(EPD_auth|EPD_username)\s*=.*$', re.M)

            with open(filename, 'r') as fi:
                data = fi.read()

            if pat.search(data):
                data = pat.sub("", data)

            with open(filename, 'w') as fo:
                fo.write(data)
        else:
            self.auth.change_auth(filename)

    def _checked_change_auth(self, auth, session, filename):
        user = {}

        session.authenticate(auth)
        self.update(auth=auth)
        self._change_auth(filename)

        user = UserInfo.from_session(session)
        print(subscription_message(self, user))
        return user

    def _set_auth(self, auth):
        """ Set the internal authentication information.

        Parameters
        ----------
        auth : Auth-like
            The authentication information. May be a (username, password)
            tuple, or an *Auth subclass.
        """
        if isinstance(auth, tuple) and len(auth) == 2:
            username, password = auth
            if username is None:
                raise InvalidConfiguration(
                    "invalid authentication arguments: "
                    "{0}:{1}".format(username, password))
            self._auth = UserPasswordAuth(username, password)
        else:
            self._auth = auth

    def _set_indexed_repositories(self, urls):
        self._indexed_repositories = tuple(fill_url(url) for url in urls)

    def _set_max_retries(self, raw_max_retries):
        try:
            max_retries = int(raw_max_retries)
        except ValueError:
            msg = "Invalid type for 'max_retries': {0!r}"
            raise InvalidConfiguration(msg.format(raw_max_retries))
        else:
            self._max_retries = max_retries

    def _set_prefix(self, prefix):
        self._prefix = os.path.normpath(abs_expanduser(prefix))

    def _set_proxy(self, proxy_string):
        self._proxy = ProxyInfo.from_string(proxy_string)

    def _set_store_url(self, url):
        p = urllib.parse.urlparse(url)
        if p.scheme.startswith(_BROOD_PREFIX):
            url = url[len(_BROOD_PREFIX):]
            self._store_kind = STORE_KIND_BROOD
        self._store_url = url

    def _set_repository_cache(self, value):
        normalized = os.path.normpath(abs_expanduser(value))
        self._repository_cache = _get_writable_local_dir(normalized)

    def _simple_attribute_set_factory(self, attribute_name):
        return lambda value: setattr(self, attribute_name, value)


def prepend_url(filename, url):
    with open(filename, 'r+') as fp:
        data = fp.read()
        pat = re.compile(r'^IndexedRepos\s*=\s*\[\s*$', re.M)
        if not pat.search(data):
            sys.exit("Error: IndexedRepos section not found")
        data = pat.sub(r"IndexedRepos = [\n  '%s'," % url, data)
        fp.seek(0)
        fp.write(data)


def print_config(config, session):
    print("Python version:", PY_VER)
    print("enstaller version:", __version__)
    print("sys.prefix:", os.path.normpath(sys.prefix))
    print("platform:", platform.platform())
    print("architecture:", platform.architecture()[0])
    print("use_webservice:", config.use_webservice)
    if config.filename is not None:
        print("config file:", config.filename)
    print("keyring backend: %s" % (_keyring_backend_name(), ))
    print("settings:")
    print("    prefix = %s" % config.prefix)
    print("    %s = %s" % ("repository_cache", config.repository_cache))
    print("    %s = %r" % ("noapp", config.noapp))
    print("    %s = %r" % ("proxy", config.proxy))
    if not config.use_webservice:
        print("    IndexedRepos:")
        for repo in config.indexed_repositories:
            print('        %r' % repo)

    user = DUMMY_USER

    if config.auth is None:
        print("No valid auth information in configuration, cannot "
              "authenticate.")
    else:
        try:
            session.authenticate(config.auth)
            user = UserInfo.from_session(session)
        except requests.exceptions.SSLError as e:
            humanize_ssl_error_and_die(e, config.store_url)
        except Exception as e:
            print(e)
    print(subscription_message(config, user))



import base64
import json
import os.path
import shutil
import tempfile

from mock import patch

from egginst.vendor.six.moves import unittest

from enstaller.auth import DUMMY_USER, UserInfo
from enstaller.auth.auth_managers import (BroodAuthenticator,
                                          BroodBearerTokenAuth,
                                          LegacyCanopyAuthManager,
                                          OldRepoAuthManager)
from enstaller.config import Configuration, write_default_config
from enstaller.session import Session
from enstaller.errors import AuthFailedError, InvalidConfiguration
from enstaller.tests.common import (DummyAuthenticator, fake_keyring,
                                    R_JSON_NOAUTH_RESP, R_JSON_AUTH_RESP)
from enstaller.vendor import requests, responses

from .._impl import UserPasswordAuth


basic_user = UserInfo(True, first_name="Jane", last_name="Doe", has_subscription=True)
free_user = UserInfo(True, first_name="John", last_name="Smith", has_subscription=False)
anon_user = UserInfo(False)
old_auth_user = DUMMY_USER


def compute_creds(username, password):
    s = "{0}:{1}".format(username, password)
    return base64.b64encode(s.encode("ascii")).rstrip()

AUTH_API_URL = 'https://api.enthought.com/accounts/user/info/'

FAKE_USER = "john.doe"
FAKE_PASSWORD = "fake_password"
FAKE_CREDS = compute_creds(FAKE_USER, FAKE_PASSWORD)


@fake_keyring
class CheckedChangeAuthTestCase(unittest.TestCase):
    def setUp(self):
        self.d = tempfile.mkdtemp()
        self.f = os.path.join(self.d, "enstaller4rc")
        self.session = Session(DummyAuthenticator(), self.d)

    def tearDown(self):
        self.session.close()
        shutil.rmtree(self.d)

    @responses.activate
    def test_no_acct(self):
        # Given
        url = "https://acme.com"
        auth_url = url + "/accounts/user/info/"

        config = Configuration()
        config.update(store_url=url)

        def callback(request):
            if auth != ("valid_user", "valid_password"):
                return (403, {}, "")
            return (200, {}, json.dumps(R_JSON_AUTH_RESP))

        responses.add_callback(responses.GET, auth_url, callback)

        write_default_config(self.f)

        with Session.from_configuration(config) as session:
            config = Configuration()
            config.update(store_url=url)

            auth = ("invalid_user", "invalid_password")

            with self.assertRaises(AuthFailedError):
                usr = config._checked_change_auth(auth, session, self.f)

            config = Configuration()
            auth = ("valid_user", "valid_password")
            usr = config._checked_change_auth(auth, session, self.f)

            self.assertTrue(usr.is_authenticated)
            self.assertEqual(config.auth,
                             UserPasswordAuth("valid_user", "valid_password"))

    def test_remote_success(self):
        write_default_config(self.f)

        config = Configuration()
        auth = ("usr", "password")
        session = Session(DummyAuthenticator(old_auth_user), self.d)

        with session:
            usr = config._checked_change_auth(auth, session, self.f)
            self.assertEqual(usr, UserInfo(True))

    def test_nones(self):
        config = Configuration()

        with self.assertRaises(InvalidConfiguration):
            config.update(auth=(None, None))


class AuthManagerBase(unittest.TestCase):
    klass = None

    def setUp(self):
        self.prefix = tempfile.mkdtemp()

        self.config = Configuration()
        self.config.update(use_webservice=False,
                           indexed_repositories=["http://acme.com"])
        self.session = Session(self.klass.from_configuration(self.config),
                               self.prefix)

    def tearDown(self):
        shutil.rmtree(self.prefix)
        self.session.close()


class TestOldReposAuthManager(AuthManagerBase):
    klass = OldRepoAuthManager

    @responses.activate
    def test_from_configuration(self):
        # Given
        responses.add(responses.HEAD, self.config.indices[0][0], status=200,
                      body=json.dumps(R_JSON_AUTH_RESP))
        authenticator = OldRepoAuthManager.from_configuration(self.config)
        session = Session(authenticator, self.prefix)

        # When
        with session:
            session.authenticate((FAKE_USER, FAKE_PASSWORD))

        # Then
        self.assertIsInstance(session._raw.auth, requests.auth.HTTPBasicAuth)
        self.assertEqual(session._raw.auth.username, FAKE_USER)
        self.assertEqual(session._raw.auth.password, FAKE_PASSWORD)

    @responses.activate
    def test_simple(self):
        # Given
        responses.add(responses.HEAD, self.config.indices[0][0], status=200,
                      body=json.dumps(R_JSON_AUTH_RESP))

        # When
        # Then no exception
        self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

    def test_connection_failure(self):
        with patch.object(self.session._raw, "head",
                          side_effect=requests.exceptions.ConnectionError):
            with self.assertRaises(AuthFailedError):
                self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

    @responses.activate
    def test_auth_failure_404(self):
        # Given
        auth = ("nono", "le petit robot")
        responses.add(responses.HEAD, self.config.indices[0][0],
                      body="", status=404,
                      content_type='application/json')

        # When/Given
        with self.assertRaises(AuthFailedError):
            self.session.authenticate(auth)

    @responses.activate
    def test_auth_failure_50x(self):
        # Given
        auth = ("nono", "le petit robot")
        responses.add(responses.HEAD, self.config.indices[0][0],
                      status=503, content_type='application/json')

        # When/Given
        with self.assertRaises(AuthFailedError):
            self.session.authenticate(auth)

    @responses.activate
    def test_auth_failure_401(self):
        # Given
        auth = ("nono", "le petit robot")
        responses.add(responses.HEAD, self.config.indices[0][0],
                      body="", status=401,
                      content_type='application/json')

        # When/Given
        with self.assertRaises(AuthFailedError):
            self.session.authenticate(auth)


class TestLegacyCanopyAuthManager(AuthManagerBase):
    klass = LegacyCanopyAuthManager

    @responses.activate
    def test_from_configuration(self):
        # Given
        responses.add(responses.GET, self.config.api_url, status=200,
                      body=json.dumps(R_JSON_AUTH_RESP))
        authenticator = LegacyCanopyAuthManager.from_configuration(self.config)
        session = Session(authenticator, self.prefix)

        # When
        with session:
            # Then no exception
            session.authenticate((FAKE_USER, FAKE_PASSWORD))

    @responses.activate
    def test_simple(self):
        # Given
        responses.add(responses.GET, self.config.api_url, status=200,
                      body=json.dumps(R_JSON_AUTH_RESP))

        # When
        # Then no exception
        self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

    def test_connection_failure(self):
        with patch.object(self.session._raw, "get",
                          side_effect=requests.exceptions.ConnectionError):
            with self.assertRaises(AuthFailedError):
                self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

    @responses.activate
    def test_http_failure(self):
        # Given
        config = Configuration()
        responses.add(responses.GET, config.api_url, body="", status=404,
                      content_type='application/json')

        # When/Then
        with self.assertRaises(AuthFailedError):
            self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

    @responses.activate
    def test_unauthenticated_user(self):
        # Given
        responses.add(responses.GET, self.config.api_url,
                      body=json.dumps(R_JSON_NOAUTH_RESP),
                      content_type='application/json')

        # When/Then
        with self.assertRaises(AuthFailedError):
            self.session.authenticate((FAKE_USER, FAKE_PASSWORD))


class TestBroodAuthManager(AuthManagerBase):
    klass = BroodAuthenticator

    def setUp(self):
        AuthManagerBase.setUp(self)
        self.token_url = self.config.store_url + "/api/v0/json/auth/tokens/auth"

    @responses.activate
    def test_from_configuration(self):
        # Given
        responses.add(responses.POST, self.token_url, status=200,
                      body=json.dumps({"token": "dummy token"}))
        authenticator = BroodAuthenticator.from_configuration(self.config)
        session = Session(authenticator, self.prefix)
        r_auth = BroodBearerTokenAuth("dummy token")

        # When
        with session:
            session.authenticate((FAKE_USER, FAKE_PASSWORD))

            # Then
            self.assertIsInstance(session._raw.auth, BroodBearerTokenAuth)
            self.assertEqual(session._raw.auth._token, r_auth._token)

    @responses.activate
    def test_simple(self):
        # Given
        responses.add(responses.POST, self.token_url, status=200,
                      body=json.dumps({"token": "dummy token"}))
        r_auth = BroodBearerTokenAuth("dummy token")

        # When
        self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

        # Then
        self.assertIsInstance(self.session._raw.auth, BroodBearerTokenAuth)
        self.assertEqual(self.session._raw.auth._token, r_auth._token)

    @responses.activate
    def test_bearer_token(self):
        # Given
        responses.add(responses.POST, self.token_url, status=200,
                      body=json.dumps({"token": "dummy_token"}))

        r_auth = BroodBearerTokenAuth("dummy_token")

        # When
        self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

        # Then
        self.assertIsInstance(self.session._raw.auth, BroodBearerTokenAuth)
        self.assertEqual(self.session._raw.auth._token, r_auth._token)

        # Given
        headers = {}

        def callback(request):
            headers.update(request.headers)
            return (200, {}, b"")
        responses.add_callback(responses.GET, "https://acme.com/fubar", callback)

        # When
        self.session.fetch("https://acme.com/fubar")

        # Then
        self.assertEqual(headers["Authorization"], "Bearer dummy_token")

    def test_connection_failure(self):
        with patch.object(self.session._raw, "post",
                          side_effect=requests.exceptions.ConnectionError):
            with self.assertRaises(AuthFailedError):
                self.session.authenticate((FAKE_USER, FAKE_PASSWORD))

    @responses.activate
    def test_http_failure(self):
        # Given
        responses.add(responses.POST, self.token_url, body="", status=403,
                      content_type='application/json')

        # When/Then
        with self.assertRaises(AuthFailedError):
            self.session.authenticate((FAKE_USER, FAKE_PASSWORD))


class TestAuthenticate(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    @fake_keyring
    @responses.activate
    def test_use_webservice_invalid_user(self):
        # Given
        config = Configuration()
        responses.add(responses.GET, config.api_url,
                      body=json.dumps(R_JSON_NOAUTH_RESP),
                      content_type='application/json')

        # When/Then
        session = Session(LegacyCanopyAuthManager(config.api_url), self.prefix)
        with session:
            with self.assertRaises(AuthFailedError):
                session.authenticate((FAKE_USER, FAKE_PASSWORD))


class SearchTestCase(unittest.TestCase):
    pass


class InstallTestCase(unittest.TestCase):
    pass



from __future__ import absolute_import, print_function

import datetime
import sys
import textwrap

from enstaller.auth import UserInfo
from enstaller.freeze import get_freeze_list
from enstaller.history import History
from enstaller.repository import Repository

from .utils import (FMT, FMT4, install_req, install_time_string,
                    name_egg, print_installed, updates_check)


def env_option(prefixes):
    print("Prefixes:")
    for p in prefixes:
        print('    %s%s' % (p, ['', ' (sys)'][p == sys.prefix]))


def freeze(prefixes):
    for package in get_freeze_list(prefixes):
        print(package)


def imports_option(repository):
    print(FMT % ('Name', 'Version', 'Location'))
    print(60 * "=")

    names = set(package.name for package in repository.iter_packages())
    for name in sorted(names, key=lambda s: s.lower()):
        packages = repository.find_packages(name)
        loc = 'sys' if packages[0].store_location == sys.prefix else 'user'
        print(FMT % (name, packages[0].full_version, loc))


def info_option(remote_repository, installed_repository, name):
    name = name.lower()
    print('Package:', name)
    print(install_time_string(installed_repository, name))
    pad = 4*' '
    for metadata in remote_repository.find_sorted_packages(name):
        print('Version: ' + metadata.full_version)
        print(pad + 'Product: %s' % metadata.product)
        print(pad + 'Available: %s' % metadata.available)
        print(pad + 'Python version: %s' % metadata.python)
        print(pad + 'Store location: %s' % metadata.store_location)
        last_mtime = datetime.datetime.fromtimestamp(metadata.mtime)
        print(pad + 'Last modified: %s' % last_mtime)
        print(pad + 'MD5: %s' % metadata.md5)
        print(pad + 'Size: %s' % metadata.size)
        reqs = set(r for r in metadata.packages)
        print(pad + "Requirements: %s" % (', '.join(sorted(reqs)) or None))


def install_from_requirements(enpkg, config, args):
    """
    Install a set of requirements specified in the requirements file.
    """
    with open(args.requirements, "r") as fp:
        for req in fp:
            args.no_deps = True
            install_req(enpkg, config, req.rstrip(), args)


def list_option(prefixes, pat=None):
    for prefix in reversed(prefixes):
        print("prefix:", prefix)
        repository = Repository._from_prefixes([prefix])
        print_installed(repository, pat)
        print()


def print_history(prefix):
    h = History(prefix)
    h.update()
    h.print_log()


def revert(enpkg, revert_arg):
    actions = enpkg.revert_actions(revert_arg)
    if actions:
        enpkg.execute(actions)
    else:
        print("Nothing to do")


def search(remote_repository, installed_repository, config, session, pat=None):
    """
    Print the packages that are available in the (remote) repository.

    It also prints a '*' in front of packages already installed.
    """
    # Flag indicating if the user received any 'not subscribed to'
    # messages
    subscribed = True

    print(FMT4 % ('Name', '  Versions', 'Product', 'Note'))
    print(80 * '=')

    names = {}
    for metadata in remote_repository.iter_packages():
        names[metadata.name] = metadata.name

    installed = {}
    for package in installed_repository.iter_packages():
        installed[package.name] = package.full_version

    for name in sorted(names, key=lambda s: s.lower()):
        if pat and not pat.search(name):
            continue
        disp_name = names[name]
        installed_version = installed.get(name)
        for metadata in remote_repository.find_sorted_packages(name):
            version = metadata.full_version
            disp_ver = (('* ' if installed_version == version else '  ') +
                        version)
            available = metadata.available
            product = metadata.product
            if not available:
                subscribed = False
            print(FMT4 % (disp_name, disp_ver, product,
                  '' if available else 'not subscribed to'))
            disp_name = ''

    if config.use_webservice and not subscribed:
        user = UserInfo.from_session(session)
        msg = textwrap.dedent("""\
            Note: some of those packages are not available at your current
            subscription level ({0!r}).""".format(user.subscription_level))
        print(msg)


def update_all(enpkg, config, args):
    updates, EPD_update = updates_check(enpkg._remote_repository,
                                        enpkg._installed_repository)
    if not (updates or EPD_update):
        print("No new version of any installed package is available")
    else:
        if EPD_update:
            new_EPD_version = EPD_update[0]['update'].full_version
            print("EPD", new_EPD_version, "is available. "
                  "To update to it (with confirmation warning), "
                  "run 'enpkg epd'.")
        if updates:
            print("The following updates and their dependencies "
                  "will be installed")
            print(FMT % ('Name', 'installed', 'available'))
            print(60 * "=")
            for update in updates:
                print(FMT % (update['current'].name,
                             update['current'].full_version,
                             update['update'].full_version))
            for update in updates:
                install_req(enpkg, config, update['current'].name, args)


def whats_new(remote_repository, installed_repository):
    updates, EPD_update = updates_check(remote_repository,
                                        installed_repository)
    if not (updates or EPD_update):
        print("No new version of any installed package is available")
    else:
        if EPD_update:
            new_EPD_version = EPD_update[0]['update'].full_version
            print("EPD", new_EPD_version, "is available. "
                  "To update to it (with confirmation warning), run "
                  "'enpkg epd'.")
        if updates:
            print(FMT % ('Name', 'installed', 'available'))
            print(60 * "=")
            for update in updates:
                print(FMT % (name_egg(update['current'].key),
                             update['current'].full_version,
                             update['update'].full_version))



from __future__ import print_function

import contextlib
import functools
import logging
import os

logger = logging.getLogger(__name__)


def slow(t):
    """
    Label a test as slow.

    Parameters
    ----------
    t : callable
        The test requiring network connectivity.

    Returns
    -------
    t : callable
        The decorated test `t`.
    """
    t.slow = True

    @functools.wraps(t)
    def slow_wrapper(*args, **kwargs):
        return t(*args, **kwargs)

    return slow_wrapper


class ControlledEnv(object):
    """
    A special os.environ that can be used for mocking os.environ.

    Beyond avoiding modifying os.environ directly, this class allows some keys
    to be ignored

    Parameters
    ----------
    ignored_keys: list
        If specified, list of keys that will be ignored.
    environ: dict
        If specified, the dictionary to use for underlying data. Default is to
        use os.environ. In both cases, the dict is copied.

    Examples
    --------
    >>> env = ControlledEnv(["USERNAME"])
    >>> "USERNAME" in env:
    False
    """

    def __init__(self, ignored_keys=None, environ=None):
        if ignored_keys is None:
            ignored_keys = set()
        self._ignored_keys = ignored_keys

        if environ is None:
            self._data = os.environ.copy()
        else:
            self._data = environ.copy()

    def __getitem__(self, name):
        if name in self._ignored_keys:
            raise KeyError("Cannot access key {0}".format(name))
        else:
            return self._data[name]

    def get(self, name, default=None):
        if name in self._data and name not in self._ignored_keys:
            return self._data[name]
        else:
            return default

    def keys(self):
        return [k for k in self._data if k not in self._ignored_keys]

    def __contains__(self, key):
        return key not in self._ignored_keys and key in self._data

    def __setitem__(self, name, value):
        self._data[name] = value


@contextlib.contextmanager
def assert_same_fs(test_case, prefix):
    all_files = []
    for root, dirs, files in os.walk(prefix):
        for f in files:
            all_files.extend(os.path.join(root, f) for f in files)
    old_state = set(all_files)

    yield

    for root, dirs, files in os.walk(prefix):
        for f in files:
            path = os.path.join(root, f)
            if path not in old_state:
                test_case.fail("Unexpected file: {0!r}".format(path))



from __future__ import print_function

import sys

import json
import time

from os.path import join

from egginst._zipfile import ZipFile
from egginst.utils import parse_assignments
from egginst.vendor.six.moves import StringIO


# Path relative to EGG-INFO in egg, or $RPPT/EGG-INFO/$package_name when
# installed
APPINST_PATH = join("inst", "appinst.dat")

SPEC_DEPEND_KEYS = ('name', 'version', 'build', 'arch', 'platform', 'osdist',
                    'python', 'packages')


def parse_rawspec(data):
    # XXX: hack to workaround 2.6-specific bug with ast-parser and
    # unicode.
    if sys.version_info < (2, 7, 3):
        spec = parse_assignments(StringIO(data.encode("utf8")))
    else:
        spec = parse_assignments(StringIO(data))
    res = {}
    for k in SPEC_DEPEND_KEYS:
        res[k] = spec[k]
    return res


def info_from_z(z):
    res = dict(type='egg')

    arcname = 'EGG-INFO/spec/depend'
    if arcname in z.namelist():
        res.update(parse_rawspec(z.read(arcname).decode("utf8")))

    arcname = 'EGG-INFO/info.json'
    if arcname in z.namelist():
        res.update(json.loads(z.read(arcname).decode("utf8")))

    res['name'] = res['name'].lower().replace('-', '_')
    return res


def create_info(egg, extra_info=None):
    info = dict(key=egg.fn)
    info.update(info_from_z(egg.z))
    info['ctime'] = time.ctime()
    # FIXME: hook kept for compat for now.
    info['hook'] = False
    if extra_info:
        info.update(extra_info)

    try:
        del info['available']
    except KeyError:
        pass

    with open(join(egg.meta_dir, '_info.json'), 'w') as fo:
        json.dump(info, fo, indent=2, sort_keys=True)

    return info


def is_custom_egg(egg):
    """
    Return True if the egg is built using Enthought build infrastructure, False
    otherwise.

    Note
    ----
    This is not 100 % reliable, as some Enthought eggs don't always have any
    specific metadata.
    """
    with ZipFile(egg) as zp:
        for dest in ("spec/depend", "inst/targets.dat"):
            try:
                zp.getinfo("EGG-INFO/{0}".format(dest))
                return True
            except KeyError:
                pass
        return False



import json

from os.path import isfile, join


_INFO_FILENAME = "_info.json"


def meta_dir_from_prefix(prefix, package_name):
    return join(prefix, "EGG-INFO", package_name)


def meta_info_from_prefix(prefix, package_name):
    return join(meta_dir_from_prefix(prefix, package_name), _INFO_FILENAME)


def info_from_metadir(meta_dir):
    path = join(meta_dir, _INFO_FILENAME)
    if isfile(path):
        try:
            with open(path) as fp:
                info = json.load(fp)
        except ValueError:
            return None
        info['installed'] = True
        info['meta_dir'] = meta_dir
        return info
    return None



import ctypes
import os.path

from ctypes.wintypes import BOOL, DWORD, HANDLE, LPWSTR
from ctypes import POINTER, c_void_p


# CreateFile ctypes implementation from SO
GENERIC_READ = 0x80000000
GENERIC_WRITE = 0x40000000

FILE_SHARE_DELETE = 0x00000004
FILE_SHARE_READ = 0x00000001
FILE_SHARE_WRITE = 0x00000002
FILE_SHARE_READ_WRITE = (FILE_SHARE_READ | FILE_SHARE_WRITE)

OPEN_EXISTING = 3

IO_REPARSE_TAG_MOUNT_POINT = 0xA0000003
REPARSE_MOUNTPOINT_HEADER_SIZE = 8

FSCTL_SET_REPARSE_POINT = 589988
FILE_FLAG_OPEN_REPARSE_POINT = 2097152
FILE_FLAG_BACKUP_SEMANTICS = 33554432
FILE_FLAG_REPARSE_BACKUP = 35651584  # FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS

INVALID_HANDLE_VALUE = -1
LPOVERLAPPED = c_void_p
LPSECURITY_ATTRIBUTES = c_void_p

NULL = 0
FALSE = BOOL(0)
TRUE = BOOL(1)


class _FileHandle(object):
    def __init__(self, handle):
        self._handle = handle

    def close(self):
        ctypes.windll.kernel32.CloseHandle(self._handle)


def CreateFile(filename, access, sharemode, attributes, creation, flags,
               template_file):
    # attributes and template_file ignored, kept for backward compat with
    # win32file
    return _FileHandle(HANDLE(ctypes.windll.kernel32.CreateFileW(
                       LPWSTR(filename),
                       DWORD(access),
                       DWORD(sharemode),
                       LPSECURITY_ATTRIBUTES(NULL),
                       DWORD(creation),
                       DWORD(flags),
                       HANDLE(NULL))))


class FILETIME(ctypes.Structure):
    _fields_ = [
        ("dwLowDateTime", DWORD),
        ("dwHighDateTime", DWORD)
    ]


class BY_HANDLE_FILE_INFORMATION(ctypes.Structure):
    _fields_ = [
        ("dwFileAttributes", DWORD),
        ("ftCreationTime", FILETIME),
        ("ftLastAccessTime", FILETIME),
        ("ftLastWriteTime", FILETIME),
        ("dwVolumeSerialNumber", DWORD),
        ("nFileSizeHigh", DWORD),
        ("nFileSizeLow", DWORD),
        ("nNumberOfLinks", DWORD),
        ("nFileIndexHigh", DWORD),
        ("nFileIndexLow", DWORD)
    ]

# http://msdn.microsoft.com/en-us/library/windows/desktop/aa364952
GetFileInformationByHandle = ctypes.windll.kernel32.GetFileInformationByHandle
GetFileInformationByHandle.argtypes = [HANDLE, POINTER(BY_HANDLE_FILE_INFORMATION)]
GetFileInformationByHandle.restype = BOOL


# samefile code adapted from
# http://timgolden.me.uk/python/win32_how_do_i/see_if_two_files_are_the_same_file.html
def _get_read_handle(filename):
    if os.path.isdir(filename):
        dwFlagsAndAttributes = FILE_FLAG_BACKUP_SEMANTICS
    else:
        dwFlagsAndAttributes = 0
    return CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, None,
                      OPEN_EXISTING, dwFlagsAndAttributes, None)


def _get_unique_id(hFile):
    file_info = BY_HANDLE_FILE_INFORMATION()
    res = GetFileInformationByHandle(hFile, ctypes.byref(file_info))
    if res == FALSE:
        raise RuntimeError("Call to GetFileInformationByHandle failed.")
    return (file_info.dwVolumeSerialNumber,
            file_info.nFileIndexHigh,
            file_info.nFileIndexLow)


def samefile(filename1, filename2):
    hFile1 = _get_read_handle(filename1)
    try:
        hFile2 = _get_read_handle(filename2)
        try:
            return _get_unique_id(hFile1._handle) == _get_unique_id(hFile2._handle)
        finally:
            hFile2.close()
    finally:
        hFile1.close()



import calendar

from email.utils import formatdate, parsedate

from datetime import datetime, timedelta


class BaseHeuristic(object):

    def warning(self):
        """
        Return a valid 1xx warning header value describing the cache adjustments.
        """
        return '110 - "Response is Stale"'

    def update_headers(self, response):
        """Update the response headers with any new headers.

        NOTE: This SHOULD always include some Warning header to
              signify that the response was cached by the client, not by way
              of the provided headers.
              return response.
        """
        return {}

    def apply(self, response):
        response.headers.update(self.update_headers(response))
        response.headers.update({'warning': self.warning()})
        return response


class OneDayCache(BaseHeuristic):
    """
    Cache the response by providing an expires 1 day in the
    future.
    """
    def update_headers(self, response):
        headers = {}

        if 'expires' not in response.headers:
            date = parsedate(response.headers['date'])
            expires = datetime(*date[:6]) + timedelta(days=1)
            headers['expires'] = formatdate(calendar.timegm(expires.timetuple()))

        return headers



from .pep386 import IrrationalVersionError, NormalizedVersion


def normalize_version_string(version_string):
    """
    Normalize the given version string to a string that can be converted to
    a NormalizedVersion.

    This function applies various special cases needed for EPD/Canopy and not
    handled in NormalizedVersion parser.

    Parameters
    ----------
    version_string: str
        The version to convert

    Returns
    -------
    normalized_version: str
        The normalized version string. Note that this is not guaranteed to be
        convertible to a NormalizedVersion
    """
    if version_string.endswith(".dev"):
        version_string += "1"
    return version_string


class PEP386WorkaroundVersion(object):
    """A version class that supports comparison, with an escape for
    versions not compatible with PEP386.

    When comparing two versions, 3 cases arise:

        * both are valid: we use the PEP386 comparison algo
        * both are invalid: we use string comparison
        * exactly one of them is valid: the valid one is always considered
          to be greather than the invalid one
    """
    @classmethod
    def from_string(cls, s):
        try:
            normalized = normalize_version_string(s)
            version = NormalizedVersion(normalized,
                                        error_on_huge_major_num=False)
            parts = version.parts
            return cls(parts)
        except IrrationalVersionError:
            parts = s.split(".")
            return cls(parts, is_worked_around=True)

    def __init__(self, parts, is_worked_around=False):
        self._parts = parts

        if is_worked_around:
            comparable_parts = tuple(parts)
        else:
            numdot = list(parts[0])
            while len(numdot) > 0 and numdot[-1] == 0:
                numdot.pop()

            prerel_parts = parts[1]
            # This is a workaround for the verlib.py implementation of
            # pep386.
            if prerel_parts[0] == "rc":
                prerel_parts = tuple(["c"] + list(prerel_parts[1:]))
            elif prerel_parts[0] == ".dev":
                # '`' sorts before 'a'. Explicit better than implicit you said
                # ?
                prerel_parts = tuple(['`'] + list(prerel_parts[1:]))

            comparable_parts = (tuple(numdot), prerel_parts, parts[2])

        self._comparable_parts = comparable_parts
        self._is_worked_around = is_worked_around

    @property
    def is_worked_around(self):
        return self._is_worked_around

    def __str__(self):
        if self._is_worked_around:
            return ".".join(self._parts)
        else:
            return NormalizedVersion.parts_to_str(self._parts)

    def __hash__(self):
        return hash(self._comparable_parts)

    def _cannot_compare(self, other):
        raise TypeError("cannot compare %s and %s"
                        % (type(self).__name__, type(other).__name__))

    def __eq__(self, other):
        if not isinstance(other, PEP386WorkaroundVersion):
            self._cannot_compare(other)
        return self._comparable_parts == other._comparable_parts

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        if not isinstance(other, PEP386WorkaroundVersion):
            self._cannot_compare(other)
        if (self._is_worked_around and other._is_worked_around) \
           or (not self._is_worked_around and not other._is_worked_around):
            return self._comparable_parts <= other._comparable_parts
        elif self._is_worked_around:
            return True
        else:
            return False

    def __lt__(self, other):
        return self.__le__(other) and self.__ne__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __gt__(self, other):
        return not self.__le__(other)
